# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for classes in local_artifacts.py."""

import json
import re
from pathlib import Path
from textwrap import dedent
from typing import Any, ClassVar, override
from unittest import mock

import debian.deb822 as deb822

import debusine.artifacts.models as data_models
from debusine.artifacts import (
    BinaryPackage,
    BinaryPackages,
    BlhcArtifact,
    DebDiffArtifact,
    DebugSymbols,
    LintianArtifact,
    LocalArtifact,
    PackageBuildLog,
    RepositoryIndex,
    SigningInputArtifact,
    SigningOutputArtifact,
    SourcePackage,
    Upload,
    WorkRequestDebugLogs,
)
from debusine.artifacts.local_artifact import (
    AutopkgtestArtifact,
    DebianSystemImageArtifact,
    DebianSystemTarballArtifact,
    NewLocalFile,
    NewRemoteFile,
    deb822dict_to_dict,
)
from debusine.artifacts.playground import ArtifactPlayground
from debusine.assets import KeyPurpose
from debusine.client.debusine import serialize_local_artifact
from debusine.client.models import FileRequest
from debusine.test import TestCase


def unregister_local_artifact_subclass(
    category: data_models.ArtifactCategory,
) -> None:
    """
    Unregister an artifact category.

    Note that the only way to register a category is to define a new subclass.
    """
    del LocalArtifact._local_artifacts_category_to_class[category]


class LocalArtifactTests(TestCase):
    """Tests for LocalArtifact class."""

    @override
    def setUp(self) -> None:
        super().setUp()

        # Register a custom artifact category, taking care that it doesn't
        # pollute other tests.
        class TestArtifact(LocalArtifact[data_models.EmptyArtifactData]):
            """Custom artifact for testing."""

            _category = data_models.ArtifactCategory.TEST

        self.addCleanup(
            unregister_local_artifact_subclass, TestArtifact._category
        )

        self.test_artifact_class = TestArtifact
        self.category = TestArtifact._category
        self.data = data_models.EmptyArtifactData()

    def test_add_local_file_in_root(self) -> None:
        """add_local_file() add the file in .files."""
        file = self.create_temporary_file()
        artifact = self.test_artifact_class._create(
            data=self.data, files=[NewLocalFile(file=file)]
        )

        self.assertEqual(artifact.files, {file.name: file})

    def test_add_local_file_in_override_name(self) -> None:
        """add_local_file() with override_name."""
        file = self.create_temporary_file()

        name = "new_name.txt"
        artifact = self.test_artifact_class._create(
            data=self.data, files=[NewLocalFile(file=file, override_name=name)]
        )

        self.assertEqual(artifact.files, {name: file})

    def test_add_local_file_artifact_base_dir_does_not_exist(self) -> None:
        """add_local_file() raise ValueError: directory does not exist."""
        file = self.create_temporary_file()
        artifact_base_dir = Path("/does/not/exist")
        with self.assertRaisesRegex(
            ValueError,
            f'"{artifact_base_dir}" does not exist or is not a directory',
        ):
            self.test_artifact_class._create(
                data=self.data,
                files=[
                    NewLocalFile(file=file, artifact_base_dir=artifact_base_dir)
                ],
            )

    def test_add_local_file_relative_to_artifact_base_dir_does_not_exist(
        self,
    ) -> None:
        """add_local_file() raise ValueError: filename does not exist."""
        artifact_base_dir = self.create_temporary_directory()
        filename = Path("README.txt")

        filename_absolute = artifact_base_dir / filename

        with self.assertRaisesRegex(
            ValueError, f'"{filename_absolute}" does not exist'
        ):
            self.test_artifact_class._create(
                data=self.data,
                files=[
                    NewLocalFile(
                        file=filename, artifact_base_dir=artifact_base_dir
                    )
                ],
            )

    def test_add_local_file_absolute_does_not_exist(self) -> None:
        """add_local_file() raise ValueError: filename does not exist."""
        empty_dir = self.create_temporary_directory()

        filename = empty_dir / "a-file-does-not-exist.txt"

        with self.assertRaisesRegex(ValueError, f'"{filename}" does not exist'):
            self.test_artifact_class._create(
                data=self.data, files=[NewLocalFile(file=filename)]
            )

    def test_add_local_file_is_not_a_file(self) -> None:
        """add_local_file() raise ValueError: expected file is a directory."""
        empty_dir = self.create_temporary_directory()

        with self.assertRaisesRegex(ValueError, f'"{empty_dir}" is not a file'):
            self.test_artifact_class._create(
                data=self.data, files=[NewLocalFile(file=empty_dir)]
            )

    def test_add_local_file_relative_to_artifact_base_dir(self) -> None:
        """add_local_file() add relative file from an absolute directory."""
        artifact_base_dir = self.create_temporary_directory()
        filename = Path("README.txt")

        (file := artifact_base_dir / filename).write_bytes(b"")

        artifact = self.test_artifact_class._create(
            data=self.data,
            files=[
                NewLocalFile(file=filename, artifact_base_dir=artifact_base_dir)
            ],
        )

        self.assertEqual(artifact.files, {'README.txt': file})

    def test_add_local_file_artifact_raise_value_error_dir_is_relative(
        self,
    ) -> None:
        """add_local_file() raise ValueError if base_dir is relative."""
        directory = Path("does-not-exist")
        file = directory / "README.txt"
        with self.assertRaisesRegex(
            ValueError, f'"{directory}" must be absolute'
        ):
            self.test_artifact_class._create(
                data=self.data,
                files=[NewLocalFile(file=file, artifact_base_dir=directory)],
            )

    def test_add_local_file_absolute_path_no_artifact_base_dir(self) -> None:
        """add_local_file() adds file. No artifact path or artifact_dir."""
        file = self.create_temporary_file()

        artifact = self.test_artifact_class._create(
            data=self.data, files=[NewLocalFile(file=file)]
        )

        self.assertEqual(artifact.files, {file.name: file})

    def test_add_local_file_duplicated_raise_value_error(self) -> None:
        """add_local_file() raise ValueError: file is already added."""
        file = self.create_temporary_file()

        with self.assertRaisesRegex(
            ValueError,
            r"^"
            + re.escape(
                f"File with the same path ({file.name}) "
                f'is already in the artifact ("{file}" and "{file}")'
            ),
        ):
            self.test_artifact_class._create(
                data=self.data,
                files=[NewLocalFile(file=file), NewLocalFile(file=file)],
            )

    def test_add_local_file_already_remote(self) -> None:
        """add_local_file() raise ValueError: remote file is already added."""
        file = self.create_temporary_file()
        file_request = FileRequest.create_from(file)

        with self.assertRaisesRegex(
            ValueError,
            "^"
            + re.escape(
                f"File with the same path ({file.name}) "
                f'is already in the artifact ("{file_request}" and "{file}")'
            ),
        ):
            self.test_artifact_class._create(
                data=self.data,
                files=[
                    NewRemoteFile(
                        path_in_artifact=file.name, file_request=file_request
                    ),
                    NewLocalFile(file=file),
                ],
            )

    def test_add_remote_file(self) -> None:
        """add_remote_file() adds a file that does not exist locally."""
        file_request = FileRequest(
            size=1, checksums={"sha256": "0" * 64}, type="file"
        )
        artifact = self.test_artifact_class._create(
            data=self.data,
            files=[
                NewRemoteFile(
                    path_in_artifact="path", file_request=file_request
                )
            ],
        )

        self.assertEqual(artifact.remote_files, {"path": file_request})

    def test_add_remote_file_already_local(self) -> None:
        """add_remote_file() raise ValueError: local file is already added."""
        file = self.create_temporary_file()
        file_request = FileRequest.create_from(file)

        with self.assertRaisesRegex(
            ValueError,
            r"^"
            + re.escape(
                f"File with the same path ({file.name}) "
                f'is already in the artifact ("{file}" and "{file_request}")'
            ),
        ):
            self.test_artifact_class._create(
                data=self.data,
                files=[
                    NewLocalFile(file=file),
                    NewRemoteFile(
                        path_in_artifact=file.name, file_request=file_request
                    ),
                ],
            )

    def test_validate_category(self) -> None:
        """Creating a new artifact validates its category."""
        self.assertRaisesRegex(
            ValueError,
            r"Invalid category: 'nonexistent'",
            WorkRequestDebugLogs,
            category="nonexistent",
            data=self.data,
            files=[],
        )

    def test_validate_files_length(self) -> None:
        """_validate_files_length return the files."""
        files = {"package1.deb": Path(), "package2.deb": Path()}
        self.assertEqual(
            LocalArtifact._validate_files_length(files, len(files)), files
        )

    def test_validate_files_length_raise_error(self) -> None:
        """_validate_files_length raise ValueError."""
        files: dict[str, Path] = {}
        expected = 1
        with self.assertRaisesRegex(
            ValueError, f"Expected number of files: {expected} Actual: 0"
        ):
            LocalArtifact._validate_files_length(files, expected)


class PackageBuildLogTests(TestCase):
    """Tests for PackageBuildLog."""

    @override
    def setUp(self) -> None:
        super().setUp()

        self.directory = self.create_temporary_directory()

        self.build_filename = "log.build"
        self.build_file = self.directory / self.build_filename
        self.build_file.write_text("A line of log")

    def test_create(self) -> None:
        """create() method return the expected PackageBuildLog."""
        source = "hello"
        version = "2.10-2"
        architecture = "amd64"

        artifact = PackageBuildLog.create(
            file=self.directory / "log.build",
            source=source,
            version=version,
            architecture=architecture,
        )

        expected_data = data_models.DebianPackageBuildLog(
            source=source,
            version=version,
            architecture=architecture,
            filename="log.build",
        )
        expected_files = {self.build_filename: self.build_file}
        expected_content_types = {
            self.build_filename: "text/plain; charset=utf-8"
        }

        expected = PackageBuildLog(
            category=PackageBuildLog._category,
            files=expected_files,
            content_types=expected_content_types,
            data=expected_data,
        )
        self.assertEqual(artifact, expected)
        self.assertIsInstance(artifact, PackageBuildLog)

    def test_validate_files_must_be_one_raise(self) -> None:
        """Raise ValueError: unexpected number of files."""
        with self.assertRaisesRegex(
            ValueError, "^Expected number of files: 1 Actual: 0$"
        ):
            PackageBuildLog.validate_files_length_is_one({})

    def test_validate_files_number_success(self) -> None:
        """files_number_must_be_one() return files."""
        files = {"log.build": self.build_file}

        self.assertEqual(
            PackageBuildLog.validate_files_length_is_one(files),
            files,
        )

    def test_validate_file_ends_with_build(self) -> None:
        """File ends with .build: files_must_end_in_build return files."""
        files = {"log.build": self.build_file}

        self.assertEqual(PackageBuildLog.file_must_end_in_build(files), files)

    def test_validate_file_ends_with_build_raise_value_error(self) -> None:
        """Raise ValueError: file does not end in .build."""
        file = self.create_temporary_file(suffix=".txt")

        expected_message = (
            rf"""^Valid file suffixes: \['\.build'\]. """
            rf"""Invalid filename: "{file.name}"$"""
        )

        with self.assertRaisesRegex(ValueError, expected_message):
            PackageBuildLog.file_must_end_in_build({file.name: file})

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        artifact = PackageBuildLog.create(
            file=self.build_file,
            source="hello",
            version="2.10-2",
            architecture="amd64",
        )

        # No exception is raised
        json.dumps(
            serialize_local_artifact(artifact, workspace="some-workspace")
        )


class UploadTests(TestCase):
    """Tests for Upload."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.workspace = "workspace"

        self.file = FileRequest(
            size=10, checksums={"sha256": "hash"}, type="file"
        )

    def test_create(self) -> None:
        """create() return the correct object."""
        directory = self.create_temporary_directory()

        changes_filename = "python-network.changes"
        changes_file = directory / changes_filename

        deb = self.create_temporary_file(
            suffix="_1.0_amd64.deb", directory=directory
        )
        dsc = self.create_temporary_file(suffix=".dsc", directory=directory)
        changes = self.write_changes_file(changes_file, [dsc, deb])

        artifact = Upload.create(changes_file=changes_file)

        files = {
            changes_filename: changes_file,
            dsc.name: dsc,
            deb.name: deb,
        }
        content_types = {changes_filename: "text/plain; charset=utf-8"}
        data = data_models.DebianUpload(changes_fields=changes, type="dpkg")

        expected = Upload(
            category=Upload._category,
            files=files,
            content_types=content_types,
            data=data,
        )

        self.assertEqual(artifact, expected)

    def test_create_no_files_in_changes(self) -> None:
        """Upload.create() return a package, .changes has no files."""
        changes_file = self.create_temporary_file(suffix=".changes")

        self.write_changes_file(changes_file, files=[])

        binary_upload = Upload.create(changes_file=changes_file)

        files = {changes_file.name: changes_file}

        self.assertEqual(binary_upload.files, files)

    def test_create_multiple_path_components(self) -> None:
        directory = self.create_temporary_directory()
        changes_file = directory / "hello.changes"
        changes_file.write_text(
            dedent("""\
                Format: 1.8
                Source: hello
                Architecture: all
                Version: 1.0
                Files:
                 0000 1 devel optional ../hello_all.deb
                Checksums-Sha1:
                 0000 1 ../hello_all.deb
                Checksums-Sha256:
                 0000 1 ../hello_all.deb
            """)
        )

        with self.assertRaisesRegex(
            ValueError,
            r"'hello\.changes' has an entry in Files with more than one path"
            r" component: '\.\./hello_all\.deb'",
        ):
            Upload.create(changes_file=changes_file)

    def test_create_allow_remote(self) -> None:
        """Upload.create() can allow referenced files to be missing locally."""
        directory = self.create_temporary_directory()
        changes_file = directory / "foo_1.0.changes"
        deb = self.create_temporary_file(
            suffix="_1.0_amd64.deb", directory=directory
        )
        dsc = self.create_temporary_file(suffix=".dsc", directory=directory)
        changes = self.write_changes_file(changes_file, [dsc, deb])
        expected = Upload(
            category=Upload._category,
            files={changes_file.name: changes_file, dsc.name: dsc},
            remote_files={deb.name: FileRequest.create_from(deb)},
            content_types={changes_file.name: "text/plain; charset=utf-8"},
            data=data_models.DebianUpload(changes_fields=changes, type="dpkg"),
        )
        deb.unlink()

        artifact = Upload.create(changes_file=changes_file, allow_remote=True)

        self.assertEqual(artifact, expected)

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        directory = self.create_temporary_directory()
        output_file = self.create_temporary_file(
            suffix=".deb", directory=directory
        )
        changes = directory / "package.changes"
        self.write_changes_file(changes, files=[output_file])
        artifact = Upload.create(changes_file=changes)

        json.dumps(
            serialize_local_artifact(artifact, workspace="some workspace")
        )

    def test_files_no_contain_changes_raise_value_error(self) -> None:
        """Raise ValueError: missing expected .changes file."""
        file = self.create_temporary_file()

        upload = Upload.model_construct(
            category=Upload._category,
            files={file.name: file},
            data=data_models.DebianUpload.model_construct(
                type="dpkg", changes_fields={}
            ),
        )

        expected_message = rf"Expecting 1 \.changes file in \['{file.name}'\]"
        with self.assertRaisesRegex(ValueError, expected_message):
            # Model validators appear as non-callable proxies while
            # type-checking, but can be called at run-time.
            upload.files_contain_changes()  # type: ignore[operator]

    def test_files_contain_2_changes_raise_value_error(self) -> None:
        """Raise ValueError: two .changes files."""
        file1 = self.create_temporary_file(suffix=".changes")
        file2 = self.create_temporary_file(suffix=".changes")

        upload = Upload.model_construct(
            category=Upload._category,
            files={
                file1.name: file1,
                file2.name: file2,
            },
            data=data_models.DebianUpload.model_construct(
                type="dpkg", changes_fields={}
            ),
        )

        expected_message = r"Expecting 1 \.changes file in \["
        with self.assertRaisesRegex(ValueError, expected_message):
            # Model validators appear as non-callable proxies while
            # type-checking, but can be called at run-time.
            upload.files_contain_changes()  # type: ignore[operator]

    def test_files_contain_changes_return_files(self) -> None:
        """Return files: .changes found."""
        changes_file = self.create_temporary_file(suffix=".changes")

        self.write_changes_file(changes_file, files=[changes_file])

        upload = Upload.model_construct(
            category=Upload._category,
            files={changes_file.name: changes_file},
            data=data_models.DebianUpload.model_construct(
                type="dpkg", changes_fields={}
            ),
        )

        # Model validators appear as non-callable proxies while
        # type-checking, but can be called at run-time.
        self.assertEqual(
            upload.files_contain_changes(),  # type: ignore[operator]
            upload,
        )

    def test_files_contains_files_in_changes_call_implementation(self) -> None:
        """Test files_contains_files_in_dsc call the utils method."""
        files = {"a": Path("b")}
        patcher = mock.patch(
            "debusine.artifacts.local_artifact.files_in_meta_file_match_files"
        )
        mocked = patcher.start()
        self.addCleanup(patcher.stop)

        upload = Upload.model_construct(
            category=Upload._category,
            files=files,
            data=data_models.DebianUpload.model_construct(
                type="dpkg", changes_fields={}
            ),
        )

        self.assertEqual(
            # Model validators appear as non-callable proxies while
            # type-checking, but can be called at run-time.
            upload.files_contains_files_in_changes(),  # type: ignore[operator]
            upload,
        )

        mocked.assert_called_with(
            ".changes",
            deb822.Changes,
            files,
        )


class Deb822DictToDictTests(TestCase):
    """Tests for deb822dict_to_dict function."""

    def assert_deb822dict_return_correct_value(
        self, value: Any, expected: Any
    ) -> None:
        """Use deb822dict_to_dict(value) and check compare with expected."""
        actual = deb822dict_to_dict(value)

        self.assertEqual(actual, expected)
        json.dumps(actual)

    def test_dict(self) -> None:
        """Python dictionary is returned as it is."""
        to_convert = {"a": "b"}
        expected = {"a": "b"}
        self.assert_deb822dict_return_correct_value(to_convert, expected)

    def test_deb822_dict(self) -> None:
        """Deb822Dict is converted to a Python dict."""
        to_convert = deb822.Deb822Dict({"a": "b"})
        expected = {"a": "b"}
        self.assert_deb822dict_return_correct_value(to_convert, expected)

    def test_list(self) -> None:
        """A list with a Deb822Dict is returned as expected."""
        to_convert = ["key", deb822.Deb822Dict({"a": "b"})]
        expected = ["key", {"a": "b"}]
        self.assert_deb822dict_return_correct_value(to_convert, expected)

    def test_str(self) -> None:
        """A string is returned as it is."""
        to_convert = deb822dict_to_dict("all")
        expected = "all"
        self.assert_deb822dict_return_correct_value(to_convert, expected)


class BinaryPackageTests(TestCase):
    """Tests for BinaryPackage."""

    def test_create(self) -> None:
        """create() return the expected package: correct data and files."""
        srcpkg_name = "hello-traditional"
        srcpkg_version = "2.10-2"
        version = "1.1.1"
        architecture = "amd64"
        directory = self.create_temporary_directory()
        file = directory / f"hello_{version}_amd64.deb"
        self.write_deb_file(
            file,
            source_name=srcpkg_name,
            source_version=srcpkg_version,
            control_file_names=["md5sums"],
        )

        artifact = BinaryPackage.create(file=file)

        expected_data = data_models.DebianBinaryPackage(
            srcpkg_name=srcpkg_name,
            srcpkg_version=srcpkg_version,
            deb_fields={
                "Package": "hello",
                "Version": version,
                "Architecture": architecture,
                "Maintainer": "Example Maintainer <example@example.org>",
                "Description": "Example description",
                "Source": f"{srcpkg_name} ({srcpkg_version})",
            },
            deb_control_files=["control", "md5sums"],
        )
        expected_files = {file.name: file}

        self.assertEqual(artifact.category, BinaryPackage._category)
        self.assertEqual(artifact.data, expected_data)
        self.assertEqual(artifact.files, expected_files)

    def test_files_must_end_in_deb_or_udeb(self) -> None:
        """files_must_end_in_deb_or_udeb() returns the files."""
        for name in ("hello_2.10-2_amd64.deb", "hello_2.10-2_amd64.udeb"):
            self.assertEqual(
                BinaryPackage.files_must_end_in_deb_or_udeb({name: Path(name)}),
                {name: Path(name)},
            )

    def test_files_must_end_in_deb_raise_value_error(self) -> None:
        """files_must_end_in_deb_or_udeb() raises ValueError."""
        filename = "README.txt"
        files = {filename: Path(filename)}
        with self.assertRaisesRegex(
            ValueError,
            rf"^Valid file suffixes: \['.deb', '.udeb'\]. "
            rf"Invalid filename: \"{filename}\"$",
        ):
            BinaryPackage.files_must_end_in_deb_or_udeb(files)

    def test_files_exactly_one(self) -> None:
        """files_exactly_one() returns the files."""
        files = {"hello.deb": Path("hello.deb")}

        self.assertEqual(BinaryPackage.files_exactly_one(files), files)

    def test_files_exactly_one_no_files(self) -> None:
        """files_exactly_one() raises ValueError: zero files."""
        with self.assertRaisesRegex(ValueError, "Must have exactly one file"):
            BinaryPackage.files_exactly_one({})

    def test_files_exactly_one_too_many_files(self) -> None:
        """files_exactly_one() raises ValueError: too many files."""
        with self.assertRaisesRegex(ValueError, "Must have exactly one file"):
            BinaryPackage.files_exactly_one(
                {
                    "hello.deb": Path("hello.deb"),
                    "hello2.deb": Path("hello2.deb"),
                }
            )

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        directory = self.create_temporary_directory()
        file = directory / "hello_2.10-2_amd64.deb"
        self.write_deb_file(file)
        artifact = BinaryPackage.create(file=file)

        # This is to verify that serialize_local_artifact() is not returning
        # some object that might raise an exception on json.dumps, e.g. if
        # it returned (as a value of the dictionary) a set().
        json.dumps(
            serialize_local_artifact(artifact, workspace="some workspace")
        )


class BinaryPackagesTests(TestCase):
    """Tests for BinaryPackages."""

    def test_create(self) -> None:
        """create() return the expected package: correct data and files."""
        srcpkg_name = "hello"
        srcpkg_version = "2.10-2"
        version = "1.1.1"
        architecture = "amd64"
        directory = self.create_temporary_directory()
        files = [
            directory / f"hello-traditional_{version}_amd64.deb",
            directory / f"hello-dbg_{version}_amd64.deb",
        ]
        for file in files:
            file.touch()

        artifact = BinaryPackages.create(
            srcpkg_name=srcpkg_name,
            srcpkg_version=srcpkg_version,
            version=version,
            architecture=architecture,
            files=files,
        )

        expected_data = data_models.DebianBinaryPackages(
            srcpkg_name=srcpkg_name,
            srcpkg_version=srcpkg_version,
            version=version,
            architecture=architecture,
            packages=["hello-traditional", "hello-dbg"],
        )
        expected_files = {
            files[0].name: files[0],
            files[1].name: files[1],
        }

        self.assertEqual(artifact.category, BinaryPackages._category)
        self.assertEqual(artifact.data, expected_data)
        self.assertEqual(artifact.files, expected_files)

    def test_files_must_end_in_deb_or_udeb(self) -> None:
        """files_must_end_in_deb_or_udeb() return the files."""
        files = {
            name: Path(name)
            for name in ("hello_2.10-2_amd64.deb", "hello_2.10-2_amd64.udeb")
        }
        self.assertEqual(
            BinaryPackages.files_must_end_in_deb_or_udeb(files),
            files,
        )

    def test_files_must_end_in_deb_raise_value_error(self) -> None:
        """files_must_end_in_deb_or_udeb() raise ValueError."""
        filename = "README.txt"
        files = {filename: Path(filename)}
        with self.assertRaisesRegex(
            ValueError,
            rf"^Valid file suffixes: \['.deb', '.udeb'\]. "
            rf"Invalid filename: \"{filename}\"$",
        ):
            BinaryPackages.files_must_end_in_deb_or_udeb(files)

    def test_files_more_than_zero(self) -> None:
        """files_more_than_zero() return the files."""
        files = {name: Path(name) for name in ("hello.deb", "hello2.deb")}

        self.assertEqual(BinaryPackages.files_more_than_zero(files), files)

    def test_files_more_than_zero_raise_value_error(self) -> None:
        """files_more_than_zero() raise ValueError: zero files."""
        with self.assertRaisesRegex(ValueError, "Must have at least one file"):
            BinaryPackages.files_more_than_zero({})

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        file = self.create_temporary_file(suffix=".deb")
        artifact = BinaryPackages.create(
            srcpkg_name="hello",
            srcpkg_version="2.10-2",
            version="2.10-2",
            architecture="amd64",
            files=[file],
        )

        # This is to verify that serialize_local_artifact() is not returning
        # some object that might raise an exception on json.dumps, e.g. if
        # it returned (as a value of the dictionary) a set().
        json.dumps(
            serialize_local_artifact(artifact, workspace="some workspace")
        )


class WorkRequestDebugLogsTests(TestCase):
    """Tests for WorkRequestDebugLogs class."""

    def test_create(self) -> None:
        """Test create(): return instance with expected files and category."""
        log_file_1 = self.create_temporary_file()
        log_file_2 = self.create_temporary_file()
        artifact = WorkRequestDebugLogs.create(files=[log_file_1, log_file_2])

        expected = WorkRequestDebugLogs(
            category=WorkRequestDebugLogs._category,
            data=data_models.EmptyArtifactData(),
            files={
                log_file_1.name: log_file_1,
                log_file_2.name: log_file_2,
            },
            content_types={
                log_file_1.name: "text/plain; charset=utf-8",
                log_file_2.name: "text/plain; charset=utf-8",
            },
        )

        self.assertEqual(artifact, expected)

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        log_file_1 = self.create_temporary_file()
        log_file_2 = self.create_temporary_file()
        artifact = WorkRequestDebugLogs.create(files=[log_file_1, log_file_2])

        # This is to verify that serialize_local_artifact() is not returning
        # some object that might raise an exception on json.dumps, e.g. if
        # it returned (as a value of the dictionary) a set().
        json.dumps(
            serialize_local_artifact(artifact, workspace="some workspace")
        )


class SourcePackageTests(TestCase):
    """Tests for SourcePackage class."""

    def test_create(self) -> None:
        """Test create(): return expected instance."""
        temp_directory = self.create_temporary_directory()
        tar_file = temp_directory / "hello_1.0.tar.gz"
        tar_file.write_bytes(b"test")
        dsc_file = temp_directory / "pkg_1.0.dsc"
        dsc_fields = ArtifactPlayground.write_deb822_file(
            deb822.Dsc, dsc_file, [tar_file], source="hello", version="1.0"
        )

        name = dsc_fields["Source"]
        version = dsc_fields["Version"]

        expected_files = {tar_file.name: tar_file, dsc_file.name: dsc_file}
        expected_data = data_models.DebianSourcePackage(
            name=name,
            version=version,
            type="dpkg",
            dsc_fields=dsc_fields,
        )

        artifact = SourcePackage.create(
            name=name, version=version, files=[tar_file, dsc_file]
        )

        self.assertEqual(artifact.category, SourcePackage._category)
        self.assertEqual(artifact.files, expected_files)
        self.assertEqual(artifact.data, expected_data)

    def test_create_invalid_dsc_raise_value_error(self) -> None:
        """Raise ValueError: invalid .dsc file."""
        file = self.create_temporary_file(suffix=".dsc")

        expected_message = rf"{file.name!r} does not contain Source field"
        with self.assertRaisesRegex(ValueError, expected_message):
            SourcePackage.create(name="hello", version="1.0", files=[file])

    def test_files_no_contain_dsc_raise_value_error(self) -> None:
        """Raise ValueError: missing expected .dsc file."""
        file = self.create_temporary_file()

        files = {file.name: file}

        expected_message = rf"Expecting 1 \.dsc file in \['{file.name}'\]"
        with self.assertRaisesRegex(ValueError, expected_message):
            SourcePackage.files_contain_one_dsc(
                files,
            )

    def test_files_contains_dsc_return_files(self) -> None:
        """Return all files: .dsc is in the files."""
        file = self.create_temporary_file(suffix=".dsc")

        files = {file.name: file}

        self.assertEqual(
            SourcePackage.files_contain_one_dsc(
                files,
            ),
            files,
        )

    def test_files_contains_files_in_dsc_call_implementation(self) -> None:
        """Test files_contains_files_in_dsc call the utils method."""
        files = {"a": Path("b")}
        patcher = mock.patch(
            "debusine.artifacts.local_artifact.files_in_meta_file_match_files"
        )
        mocked = patcher.start()
        self.addCleanup(patcher.stop)

        self.assertEqual(
            SourcePackage.files_contains_files_in_dsc(files), files
        )

        mocked.assert_called_with(".dsc", deb822.Dsc, files)

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        dir_ = self.create_temporary_directory()
        dsc_file = self.create_temporary_file(suffix=".dsc", directory=dir_)
        dsc_fields = self.write_dsc_example_file(dsc_file)
        name = dsc_fields["Source"]
        version = dsc_fields["Version"]
        files = [dsc_file]
        for file in dsc_fields["Files"]:
            (dir_ / file["name"]).write_bytes(b"")
            files.append(dir_ / file["name"])
        package = SourcePackage.create(name=name, version=version, files=files)

        # This is to verify that serialize_local_artifact() is not returning
        # some object that might raise an exception on json.dumps, e.g. if
        # it returned (as a value of the dictionary) a set().
        json.dumps(
            serialize_local_artifact(package, workspace="some workspace")
        )


class LintianTests(TestCase):
    """Tests for the LintianArtifact."""

    output_content: ClassVar[bytes]
    summary: ClassVar[data_models.DebianLintianSummary]
    analysis_content: ClassVar[bytes]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.output_content = (
            b"W: python-ping3 source: missing-license-"
            b"paragraph-in-dep5-copyright gpl-3 [debian/copyright:33]\n"
            b"P: python-ping3 source: maintainer-manual-page [debian/ping3.1]\n"
        )
        cls.summary = data_models.DebianLintianSummary(
            tags_count_by_severity={
                data_models.DebianLintianSeverity.ERROR: 1,
                data_models.DebianLintianSeverity.WARNING: 1,
            },
            package_filename={
                "binutils": "binutils_1.0.dsc",
                "cynthiune.app": "cynthiune.app_1.0_amd64.deb",
            },
            tags_found=["license-problem-convert-utf-code", "vcs-obsolete"],
            overridden_tags_found=[],
            lintian_version="1.0.0",
            distribution="bookworm",
        )
        cls.analysis_content = json.dumps(
            {
                "tags": [
                    {
                        "package": "cynthiune.app",
                        "severity": "warning",
                        "tag": "vcs-obsolete",
                        "note": "",
                        "pointer": "",
                        "explanation": "",
                        "comment": "",
                    },
                    {
                        "package": "binutils",
                        "severity": "error",
                        "tag": "license-problem-convert-utf-code",
                        "note": "Cannot ls",
                        "pointer": "src/ls.c",
                        "explanation": "",
                        "comment": "",
                    },
                ]
            }
        ).encode("utf-8")

    def test_create(self) -> None:
        """Test create(): return expected class with the files."""
        output = self.create_temporary_file(
            suffix=".txt", contents=self.output_content
        )
        analysis = self.create_temporary_file(
            suffix=".json", contents=self.analysis_content
        )

        package = LintianArtifact.create(
            lintian_output=output,
            analysis=analysis,
            architecture="i386",
            summary=self.summary,
        )

        self.assertEqual(package._category, LintianArtifact._category)
        self.assertEqual(
            package.data,
            data_models.DebianLintian(
                architecture="i386", summary=self.summary
            ),
        )
        self.assertEqual(
            package.files,
            {
                "lintian.txt": output,
                "analysis.json": analysis,
            },
        )

    def test_validate_file_analysis_is_json_valid(self) -> None:
        """Test validate_file_analysis_is_json: return files."""
        analysis = self.create_temporary_file(contents=b"{}")
        files = {"analysis.json": analysis}

        self.assertEqual(
            LintianArtifact._validate_file_analysis_is_json(files),
            files,
        )

    def test_validate_file_analysis_json_raise_value_error(self) -> None:
        """Test validate_file_analysis_is_json: raise ValueError."""
        analysis = self.create_temporary_file(contents=b":")  # invalid

        msg = "^analysis.json is not valid JSON:"
        with self.assertRaisesRegex(ValueError, msg):
            LintianArtifact._validate_file_analysis_is_json(
                {"analysis.json": analysis}
            )

    def test_validate_required_files_valid(self) -> None:
        """Test validate_required_files: is valid. return files."""
        files = {name: Path(name) for name in ("analysis.json", "lintian.txt")}

        self.assertEqual(LintianArtifact._validate_required_files(files), files)

    def test_validate_required_files_missing_file_raise_value_error(
        self,
    ) -> None:
        """Test validate_required_files: not valid, raise ValueError."""
        msg = r"^Files required: \['analysis.json', 'lintian.txt'\]$"

        with self.assertRaisesRegex(ValueError, msg):
            LintianArtifact._validate_required_files(
                {"summary.json": Path("summary.json")}
            )

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        output = self.create_temporary_file(
            suffix=".txt", contents=self.output_content
        )
        analysis = self.create_temporary_file(
            suffix=".json", contents=self.analysis_content
        )
        artifact = LintianArtifact.create(
            lintian_output=output,
            analysis=analysis,
            architecture="source",
            summary=self.summary,
        )

        # This is to verify that serialize_local_artifact() is not returning
        # some object that might raise an exception on json.dumps, e.g. if
        # it returned (as a value of the dictionary) a set().
        json.dumps(
            serialize_local_artifact(artifact, workspace="some workspace")
        )


class AutopkgtestArtifactTests(TestCase):
    """Tests for AutopkgtestArtifact class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.data_contents: dict[str, Any] = {
            "results": {
                "test": {
                    "status": "PASS",
                    "details": "test details",
                },
            },
            "cmdline": "/bin/true",
            "source_package": {
                "name": "hello",
                "version": "1.0",
                "url": "https://example.org/hello.deb",
            },
            "architecture": "amd64",
            "distribution": "debian:sid",
        }

    def test_create(self) -> None:
        """Test _create()."""
        directory = self.create_temporary_directory()

        # Create files in directory and subdirectories: to be included in
        # the package
        (summary_file := directory / "summary").write_text("summary content")

        (subdir := directory / "subdir").mkdir()
        (file_in_subdir := subdir / "some-file.txt").write_text("test")

        (binaries_subdir := directory / "binaries").mkdir()

        # Files in "binaries/" are not part of the artifact
        (binaries_subdir / "log.txt").write_text("log file")
        (binaries_subdir / "pkg.deb").write_text("deb")

        package = AutopkgtestArtifact.create(
            directory, data_models.DebianAutopkgtest(**self.data_contents)
        )

        self.assertIsInstance(package, AutopkgtestArtifact)
        self.assertEqual(
            package.data, data_models.DebianAutopkgtest(**self.data_contents)
        )

        self.assertEqual(
            package.files,
            {
                "subdir/some-file.txt": file_in_subdir,
                "summary": summary_file,
            },
        )

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        directory = self.create_temporary_directory()
        (directory / "summary").write_text("summary content")

        artifact = AutopkgtestArtifact.create(
            directory, data_models.DebianAutopkgtest(**self.data_contents)
        )

        # This is to verify that serialize_local_artifact() is not returning
        # some object that might raise an exception on json.dumps, e.g. if
        # it returned (as a value of the dictionary) a set().
        json.dumps(
            serialize_local_artifact(artifact, workspace="some workspace")
        )


class DebianSystemTarballArtifactTests(TestCase):
    """Tests for DebianSystemTarballArtifactTests."""

    def test_create(self) -> None:
        """Test _create()."""
        system_tar_xz = self.create_temporary_file(
            suffix=".tar.xz", contents=b"some-file-contents"
        )

        artifact = DebianSystemTarballArtifact.create(
            system_tar_xz,
            data={
                "vendor": "debian",
                "codename": "bookworm",
                "mirror": "https://deb.debian.org",
                "components": ["main"],
                "variant": "minbase",
                "pkglist": {},
                "architecture": "amd64",
                "with_dev": True,
                "with_init": True,
            },
        )

        self.assertEqual(artifact.files, {system_tar_xz.name: system_tar_xz})
        self.assertEqual(artifact.data.filename, system_tar_xz.name)

    def test_validate_file_name_ends_in_tar_xz(self) -> None:
        """validate_file_name_ends_in_tar_xz is valid: return files."""
        system_tar_xz = self.create_temporary_file(
            suffix=".tar.xz", contents=b"some-file-contents"
        )

        files = {system_tar_xz.name: system_tar_xz}

        self.assertEqual(
            DebianSystemTarballArtifact._validate_file_name_ends_in_tar_xz(
                files
            ),
            files,
        )

    def test_validate_file_name_ends_in_tar_xz_two_files_raise(self) -> None:
        """validate_file_name_ends_in_tar_xz raise ValueError with two files."""
        file1 = self.create_temporary_file(
            suffix="temp", contents=b"some-file-contents"
        )

        file2 = self.create_temporary_file(
            suffix="temp", contents=b"some-file-contents"
        )

        files = {file1.name: file1, file2.name: file2}

        msg = "DebianSystemTarballArtifact does not contain exactly one file"
        with self.assertRaisesRegex(ValueError, msg):
            DebianSystemTarballArtifact._validate_file_name_ends_in_tar_xz(
                files
            )

    def test_validate_file_name_ends_in_tar_xz_not_valid_raise(self) -> None:
        """validate_file_name_ends_in_tar_xz raise ValueError."""
        file = self.create_temporary_file(
            suffix="temp", contents=b"some-file-contents"
        )

        files = {file.name: file}

        msg = f"Invalid file name: '{file.name}'. Expected .tar.xz"
        with self.assertRaisesRegex(ValueError, msg):
            DebianSystemTarballArtifact._validate_file_name_ends_in_tar_xz(
                files
            )


class BlhcArtifactTests(TestCase):
    """Tests for BlhcArtifact class."""

    output_content: ClassVar[bytes]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.output_content = (
            b"LDFLAGS missing (-fPIE -pie): /usr/bin/c++  -g -O2 -fstack-"
            b"protector-strong -Wformat -Werror=format-security -Wdate-time "
            b"-D_FORTIFY_SOURCE=2  -Wl,-z,relro printf-test.cc.o "
            b"-o ../bin/printf-test\n"
        )

    def test_create(self) -> None:
        """Test _create()."""
        build_log = data_models.DebianPackageBuildLog(
            source="hello",
            version="1.0",
            architecture="amd64",
            filename="hello_1.0_amd64.buildlog",
        )
        output = self.create_temporary_file(
            suffix=".txt", contents=self.output_content
        )

        package = BlhcArtifact.create(blhc_output=output, build_log=build_log)

        self.assertIsInstance(package, BlhcArtifact)
        self.assertEqual(
            package.data,
            data_models.DebianBlhc(
                source="hello", version="1.0", architecture="amd64"
            ),
        )

        self.assertEqual(
            package.files,
            {
                "blhc.txt": output,
            },
        )

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        """upload_artifact() and relation_create() is called."""
        build_log = data_models.DebianPackageBuildLog(
            source="hello",
            version="1.0",
            architecture="amd64",
            filename="hello_1.0_amd64.buildlog",
        )

        # Create file for the artifact
        blhc_output = self.create_temporary_file(
            suffix=".txt", contents=b"NONVERBOSE BUILD: Compiling gkrust\n"
        )

        artifact = BlhcArtifact.create(
            blhc_output=blhc_output, build_log=build_log
        )

        # No exception is raised
        json.dumps(
            serialize_local_artifact(artifact, workspace="some-workspace")
        )


class DebDiffArtifactTests(TestCase):
    """Tests for DebDiffArtifact class."""

    output_content: ClassVar[bytes]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.output_content = (
            b"diff -Nru a/debian/patches/series b/debian/patches/series"
            b"--- a/debian/patches/series       2024-03-23 10:37:14.00000 +0100"
            b"+++ b/debian/patches/series       2024-08-22 12:58:30.00000 +0200"
            b"@@ -5,3 +5,4 @@"
            b" fix-ftbfs-with-gcc-12.patch"
            b" fix-aptitude-changelog-parser.patch"
            b" fix-ftbfs-with-t64.patch"
            b"+0008-Add-missing-include-to-build-with-gcc-14.patch"
        )

    def test_create(self) -> None:
        """Test _create()."""
        output = self.create_temporary_file(
            suffix=".txt", contents=self.output_content
        )

        artifact = DebDiffArtifact.create(
            debdiff_output=output, original="original", new="new", label="O → N"
        )

        self.assertIsInstance(artifact, DebDiffArtifact)
        self.assertEqual(
            artifact.data,
            data_models.DebDiff(original="original", new="new", label="O → N"),
        )

        self.assertEqual(
            artifact.files,
            {
                "debdiff.txt": output,
            },
        )

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        # Create file for the artifact
        debdiff_output = self.create_temporary_file(
            suffix=".txt", contents=b"debdiff a.deb b.deb\n"
        )

        artifact = DebDiffArtifact.create(
            debdiff_output=debdiff_output, original="original", new="new"
        )

        # No exception is raised
        json.dumps(
            serialize_local_artifact(artifact, workspace="some-workspace")
        )


class DebianSystemImageArtifactTests(TestCase):
    """Tests for DebianSystemImageArtifactTests."""

    def test_create_tar_xz(self) -> None:
        """Test _create() with .tar.xz."""
        system_tar_xz = self.create_temporary_file(
            suffix=".tar.xz", contents=b"some-file-contents"
        )

        artifact = DebianSystemImageArtifact.create(
            system_tar_xz,
            {
                "image_format": "raw",
                "boot_mechanism": "efi",
                "vendor": "debian",
                "codename": "sid",
                "mirror": "https://deb.debian.org",
                "components": ["main"],
                "filesystem": "ext4",
                "size": 12345,
                "pkglist": {},
                "architecture": "amd64",
                "with_dev": True,
                "with_init": True,
            },
        )

        self.assertEqual(artifact.files, {system_tar_xz.name: system_tar_xz})
        self.assertEqual(artifact.data.filename, system_tar_xz.name)

    def test_create_qcow2(self) -> None:
        """Test _create() with .qcow2."""
        system_qcow2 = self.create_temporary_file(
            suffix=".qcow2", contents=b"some-file-contents"
        )

        artifact = DebianSystemImageArtifact.create(
            system_qcow2,
            {
                "image_format": "qcow2",
                "boot_mechanism": "efi",
                "vendor": "debian",
                "codename": "sid",
                "mirror": "https://deb.debian.org",
                "components": ["main"],
                "filesystem": "ext4",
                "size": 12345,
                "pkglist": {},
                "architecture": "amd64",
                "with_dev": True,
                "with_init": True,
            },
        )

        self.assertEqual(artifact.files, {system_qcow2.name: system_qcow2})
        self.assertEqual(artifact.data.filename, system_qcow2.name)

    def test_validate_file_name_ending_tar_xz(self) -> None:
        """validate_files is .tar.xz: return files."""
        system_tar_xz = self.create_temporary_file(
            suffix=".tar.xz", contents=b"some-file-contents"
        )

        files = {system_tar_xz.name: system_tar_xz}

        self.assertEqual(
            DebianSystemImageArtifact._validate_files(files),
            files,
        )

    def test_validate_file_name_ending_qcow2(self) -> None:
        """validate_files is .qcow2: return files."""
        system_qcow2 = self.create_temporary_file(
            suffix=".qcow2", contents=b"some-file-contents"
        )

        files = {system_qcow2.name: system_qcow2}

        self.assertEqual(
            DebianSystemImageArtifact._validate_files(files),
            files,
        )

    def test_validate_two_files(self) -> None:
        """validate_files raise ValueError if not one file."""
        file1 = self.create_temporary_file(
            suffix="temp", contents=b"some-file-contents"
        )

        file2 = self.create_temporary_file(
            suffix="temp", contents=b"some-file-contents"
        )

        files = {file1.name: file1, file2.name: file2}

        msg = "DebianSystemImageArtifact does not contain exactly one file"
        with self.assertRaisesRegex(ValueError, msg):
            DebianSystemImageArtifact._validate_files(files)

    def test_validate_file_name_ending_not_valid_raise(self) -> None:
        """validate_files raise ValueError with other file ending."""
        file = self.create_temporary_file(
            suffix="temp", contents=b"some-file-contents"
        )

        files = {file.name: file}

        msg = f"Invalid file name: '{file.name}'. Expected .tar.xz or qcow2"
        with self.assertRaisesRegex(ValueError, msg):
            DebianSystemImageArtifact._validate_files(files)


class SigningInputArtifactTests(TestCase):
    """Tests for SigningInputArtifact."""

    def test_create(self) -> None:
        """Create a valid SigningInputArtifact."""
        base_dir = self.create_temporary_directory()
        unsigned = base_dir / "package" / "dir" / "file"
        unsigned.parent.mkdir(parents=True)
        unsigned.write_text("contents")

        artifact = SigningInputArtifact.create([unsigned], base_dir)

        self.assertEqual(
            artifact.data,
            data_models.DebusineSigningInput(
                trusted_certs=None, binary_package_name=None
            ),
        )
        self.assertEqual(artifact.files, {"package/dir/file": unsigned})
        self.assertEqual(
            artifact.content_types,
            {"package/dir/file": "application/octet-stream"},
        )

    def test_create_trusted_certs(self) -> None:
        """Create a SigningInputArtifact with trusted_certs."""
        base_dir = self.create_temporary_directory()
        unsigned_1 = base_dir / "package" / "dir" / "file"
        unsigned_1.parent.mkdir(parents=True)
        unsigned_1.write_text("contents")
        unsigned_2 = base_dir / "package" / "another-dir" / "another-file"
        unsigned_2.parent.mkdir(parents=True)
        unsigned_2.write_text("other-contents")

        artifact = SigningInputArtifact.create(
            [unsigned_1, unsigned_2], base_dir, trusted_certs=["0" * 64]
        )

        self.assertEqual(
            artifact.data,
            data_models.DebusineSigningInput(
                trusted_certs=["0" * 64], binary_package_name=None
            ),
        )
        self.assertEqual(
            artifact.files,
            {
                "package/dir/file": unsigned_1,
                "package/another-dir/another-file": unsigned_2,
            },
        )
        self.assertEqual(
            artifact.content_types,
            {
                "package/dir/file": "application/octet-stream",
                "package/another-dir/another-file": "application/octet-stream",
            },
        )

    def test_create_binary_package_name(self) -> None:
        """Create a valid SigningInputArtifact with a binary package name."""
        base_dir = self.create_temporary_directory()
        unsigned = base_dir / "package" / "dir" / "file"
        unsigned.parent.mkdir(parents=True)
        unsigned.write_text("contents")

        artifact = SigningInputArtifact.create(
            [unsigned], base_dir, binary_package_name="hello"
        )

        self.assertEqual(
            artifact.data,
            data_models.DebusineSigningInput(
                trusted_certs=None, binary_package_name="hello"
            ),
        )
        self.assertEqual(artifact.files, {"package/dir/file": unsigned})
        self.assertEqual(
            artifact.content_types,
            {"package/dir/file": "application/octet-stream"},
        )

    def test_create_no_files(self) -> None:
        """A SigningInputArtifact must have at least one file."""
        base_dir = self.create_temporary_directory()

        with self.assertRaisesRegex(ValueError, "Expected at least one file"):
            SigningInputArtifact.create([], base_dir)


class SigningOutputArtifactTests(TestCase):
    """Tests for SigningOutputArtifact."""

    def test_create(self) -> None:
        """Create a valid SigningOutputArtifact."""
        base_dir = self.create_temporary_directory()
        output_file = base_dir / "package" / "dir" / "file.sig"
        output_file.parent.mkdir(parents=True)
        output_file.write_text("signature")
        result = data_models.SigningResult(
            file="package/dir/file", output_file="package/dir/file.sig"
        )

        artifact = SigningOutputArtifact.create(
            KeyPurpose.UEFI,
            "0" * 64,
            [result],
            [output_file],
            base_dir,
        )

        self.assertEqual(
            artifact.data,
            data_models.DebusineSigningOutput(
                purpose=KeyPurpose.UEFI,
                fingerprint="0" * 64,
                results=[result],
                binary_package_name=None,
            ),
        )
        self.assertEqual(artifact.files, {"package/dir/file.sig": output_file})
        self.assertEqual(
            artifact.content_types,
            {"package/dir/file.sig": "application/octet-stream"},
        )

    def test_create_binary_package_name(self) -> None:
        """Create a valid SigningOutputArtifact with a binary package name."""
        base_dir = self.create_temporary_directory()
        output_file = base_dir / "package" / "dir" / "file.sig"
        output_file.parent.mkdir(parents=True)
        output_file.write_text("signature")
        result = data_models.SigningResult(
            file="package/dir/file", output_file="package/dir/file.sig"
        )

        artifact = SigningOutputArtifact.create(
            KeyPurpose.UEFI,
            "0" * 64,
            [result],
            [output_file],
            base_dir,
            binary_package_name="hello",
        )

        self.assertEqual(
            artifact.data,
            data_models.DebusineSigningOutput(
                purpose=KeyPurpose.UEFI,
                fingerprint="0" * 64,
                results=[result],
                binary_package_name="hello",
            ),
        )
        self.assertEqual(artifact.files, {"package/dir/file.sig": output_file})
        self.assertEqual(
            artifact.content_types,
            {"package/dir/file.sig": "application/octet-stream"},
        )


class RepositoryIndexTests(TestCase):
    """Tests for RepositoryIndex."""

    def test_create(self) -> None:
        """Create a valid RepositoryIndex."""
        directory = self.create_temporary_directory()
        (
            release := directory / "deb.debian.org_debian_dists_sid_Release"
        ).write_text("A Release file\n")

        artifact = RepositoryIndex.create(file=release, path="Release")

        self.assertEqual(
            artifact,
            RepositoryIndex(
                category=RepositoryIndex._category,
                files={"Release": release},
                data=data_models.DebianRepositoryIndex(path="Release"),
            ),
        )

    def test_create_with_content_type(self) -> None:
        """Create a RepositoryIndex with a specific Content-Type."""
        directory = self.create_temporary_directory()
        (
            release := directory / "deb.debian.org_debian_dists_sid_Release"
        ).write_text("A Release file\n")

        artifact = RepositoryIndex.create(
            file=release,
            path="Release",
            content_type="text/plain; charset=utf-8",
        )

        self.assertEqual(
            artifact,
            RepositoryIndex(
                category=RepositoryIndex._category,
                files={"Release": release},
                content_types={"Release": "text/plain; charset=utf-8"},
                data=data_models.DebianRepositoryIndex(path="Release"),
            ),
        )


class DebugSymbolsTests(TestCase):
    """Tests for DebugSymbols."""

    def test_create(self) -> None:
        """create() returns the expected
        artifact with correct data and files.
        """
        build_id_1 = "a" * 40
        build_id_2 = "b" * 40

        file_1 = self.create_temporary_file()
        file_2 = self.create_temporary_file()

        artifact = DebugSymbols.create(
            build_id_files={build_id_1: file_1, build_id_2: file_2}
        )

        self.assertEqual(artifact.category, DebugSymbols._category)
        self.assertEqual(
            artifact.data,
            data_models.DebianDebugSymbols(build_ids=[build_id_1, build_id_2]),
        )
        self.assertEqual(
            artifact.files,
            {build_id_1: file_1, build_id_2: file_2},
        )

    def test_create_empty(self) -> None:
        """create() with no files produces an artifact with empty build_ids."""
        artifact = DebugSymbols.create(build_id_files={})

        self.assertEqual(artifact.data.build_ids, [])
        self.assertEqual(artifact.files, {})

    def test_files_mismatch_raises(self) -> None:
        """Raise ValueError if files keys don't match build_ids."""
        build_id = "a" * 40
        file_ = self.create_temporary_file()

        with self.assertRaisesRegex(ValueError, "build_ids vs files mismatch"):
            DebugSymbols(
                category=DebugSymbols._category,
                data=data_models.DebianDebugSymbols(
                    build_ids=[build_id, "b" * 40]
                ),
                files={build_id: file_},
            )

    def test_extra_file_raises(self) -> None:
        """Raise ValueError if files has a key not in build_ids."""
        build_id = "a" * 40
        file_ = self.create_temporary_file()

        with self.assertRaisesRegex(ValueError, "build_ids vs files mismatch"):
            DebugSymbols(
                category=DebugSymbols._category,
                data=data_models.DebianDebugSymbols(build_ids=[]),
                files={build_id: file_},
            )

    def test_json_serializable(self) -> None:
        """Artifact returned by create is JSON serializable."""
        build_id = "a" * 40
        file_ = self.create_temporary_file()

        artifact = DebugSymbols.create(build_id_files={build_id: file_})

        json.dumps(
            serialize_local_artifact(artifact, workspace="some-workspace")
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""blhc workflow."""

from collections.abc import Callable, Sequence
from contextlib import suppress
from operator import itemgetter
from typing import Any, override

from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    DebianBlhc,
    DebianPackageBuildLog,
)
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.models import Artifact, ArtifactRelation, WorkRequest
from debusine.db.models.lookups import LookupResult
from debusine.server.collections.lookup import artifact_ids
from debusine.server.workflows import workflow_utils
from debusine.server.workflows.models import (
    BlhcWorkflowData,
    BlhcWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.regression_tracking import (
    RegressionTrackingWorkflow,
)
from debusine.server.workflows.workflow_utils import lookup_result_architectures
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BlhcData,
    BlhcFlags,
    BlhcInput,
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    OutputData,
    RegressionAnalysis,
)


class BlhcWorkflow(
    RegressionTrackingWorkflow[BlhcWorkflowData, BlhcWorkflowDynamicData]
):
    """Blhc workflow."""

    TASK_NAME = "blhc"

    def _lookup_package_build_logs(self) -> Sequence[LookupResult]:
        if self.data.binary_artifacts is not None:
            binary_artifacts = [
                result.artifact
                for result in self.work_request.lookup_multiple(
                    self.data.binary_artifacts,
                    expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
                )
                if result.artifact is not None
            ]
            # Ideally we'd have a way to express this sort of following of
            # relations in the lookup syntax.  In the meantime, synthesize
            # basic lookup results.
            return [
                LookupResult(
                    result_type=LookupResultType.ARTIFACT,
                    lookup=f"{artifact.id}@artifacts",
                    artifact=artifact,
                )
                for artifact in Artifact.objects.filter(
                    category=ArtifactCategory.PACKAGE_BUILD_LOG,
                    relations__target__in=binary_artifacts,
                    relations__type=ArtifactRelation.Relations.RELATES_TO,
                )
            ]
        else:
            # Checked by a model validator.
            assert self.data.package_build_logs is not None
            return self.work_request.lookup_multiple(
                self.data.package_build_logs,
                expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
            )

    def _get_source_package_name_and_version(
        self, result: LookupResult
    ) -> tuple[str, str]:
        if result.artifact is not None and (
            result.artifact.category == ArtifactCategory.PACKAGE_BUILD_LOG
        ):
            package_build_log_data = result.artifact.create_data()
            assert isinstance(package_build_log_data, DebianPackageBuildLog)
            return package_build_log_data.source, package_build_log_data.version
        elif (
            result.collection_item is not None
            and (result.collection_item.category == BareDataCategory.PROMISE)
            and (data := result.collection_item.data)
            and isinstance((name := data.get("source_package_name")), str)
            and isinstance((version := data.get("source_package_version")), str)
        ):
            return name, version
        else:
            raise ValueError(
                f"Unexpected result: must be either "
                f"{ArtifactCategory.PACKAGE_BUILD_LOG} artifact or promise "
                f"with 'source_package_name' and 'source_package_version'"
            )

    @override
    def compute_dynamic_data(self) -> BlhcWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: distinct ``source`` field (separated by spaces) from each of
          the artifacts' data
        :source_artifact_id: id of the source_artifact or None
        :package_build_logs_ids: ids of the package_build_logs
        """
        package_build_logs = self._lookup_package_build_logs()
        source_package_names = workflow_utils.get_source_package_names(
            package_build_logs,
            configuration_key="package_build_logs",
            artifact_expected_categories=(ArtifactCategory.PACKAGE_BUILD_LOG,),
        )

        source_artifact_id = None
        if self.data.source_artifact is not None:
            source_artifact_id = workflow_utils.source_package(self).id

        return BlhcWorkflowDynamicData(
            subject=" ".join(source_package_names),
            source_artifact_id=source_artifact_id,
            binary_artifacts_ids=(
                None
                if self.data.binary_artifacts is None
                else artifact_ids(
                    self.work_request.lookup_multiple(
                        self.data.binary_artifacts,
                        expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
                    )
                )
            ),
            package_build_logs_ids=sorted(artifact_ids(package_build_logs)),
        )

    def _has_current_reference_qa_result(
        self, package_build_log: LookupResult, architecture: str
    ) -> bool:
        """
        Return True iff we have a current reference QA result.

        A blhc analysis is outdated if the underlying source package is
        outdated (i.e. has a smaller version number) compared to what's
        available in the ``debian:suite`` collection.  The comparison needs
        to be performed based on the metadata of the linked
        :artifact:`debian:package-build-log` artifact.  Otherwise, it is
        current.
        """
        # This method is only called when update_qa_results is "yes", in
        # which case this is checked by a model validator.
        assert self.reference_qa_results is not None

        source_name, source_version = self._get_source_package_name_and_version(
            package_build_log
        )
        latest_result = self.reference_qa_results.manager.lookup(
            f"latest:blhc_{source_name}_{architecture}"
        )
        return (
            latest_result is not None
            and latest_result.data["version"] == source_version
        )

    @override
    def populate(self) -> None:
        """Create work requests."""
        environment = f"{self.data.vendor}/match:codename={self.data.codename}"
        items: list[tuple[str, LookupResult]] = []
        for package_build_log in self._lookup_package_build_logs():
            architectures = lookup_result_architectures(package_build_log)
            assert len(architectures) == 1
            architecture = architectures.pop()
            items.append((architecture, package_build_log))

        for architecture, package_build_log in sorted(items, key=itemgetter(0)):
            self._populate_single(
                package_build_log=package_build_log,
                architecture=architecture,
                extra_flags=self.data.extra_flags,
                environment=environment,
            )
        self._populate_regression_analysis_callback(
            {architecture for architecture, _ in items}
        )

    def _populate_single(
        self,
        *,
        package_build_log: LookupResult,
        architecture: str,
        extra_flags: list[BlhcFlags],
        environment: str,
    ) -> None:
        if (
            self.data.update_qa_results == RegressionTrackingUpdateQAResults.YES
            and self._has_current_reference_qa_result(
                package_build_log, architecture
            )
        ):
            return

        package_build_log_lookup = package_build_log.reconstruct()
        workflow_data_step = f"blhc-{architecture}"
        workflow_data_kwargs: dict[str, Any] = {}
        if self.data.update_qa_results != RegressionTrackingUpdateQAResults.NO:
            # When updating reference results for regression tracking, task
            # failures never cause the parent workflow or dependent tasks to
            # fail.
            workflow_data_kwargs["allow_failure"] = True
        wr, created = self.work_request_ensure_child_worker(
            task_name="blhc",
            task_data=BlhcData(
                input=BlhcInput(artifact=package_build_log_lookup),
                extra_flags=extra_flags,
                build_architecture=(
                    self.data.arch_all_build_architecture
                    if architecture == "all"
                    else architecture
                ),
                environment=environment,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"build log hardening check for {architecture}",
                step=workflow_data_step,
                **workflow_data_kwargs,
            ),
        )
        self.requires_artifact(wr, package_build_log_lookup)
        promise_name = f"{self.data.prefix}blhc-{architecture}"
        self.provides_artifact(wr, ArtifactCategory.BLHC, promise_name)

        if created and (
            self.data.update_qa_results != RegressionTrackingUpdateQAResults.NO
        ):
            # Checked by a model validator.
            assert self.data.reference_qa_results is not None

            source_name, source_version = (
                self._get_source_package_name_and_version(package_build_log)
            )

            # Back off if another workflow gets there first.
            self.skip_if_qa_result_changed(
                wr,
                package=source_name,
                architecture=architecture,
                promise_name=promise_name,
            )

            # Record results in the reference collection.
            action = ActionUpdateCollectionWithArtifacts(
                collection=self.data.reference_qa_results,
                variables={
                    "package": source_name,
                    "version": source_version,
                    "architecture": architecture,
                    "timestamp": int(
                        (self.qa_suite_changed or timezone.now()).timestamp()
                    ),
                    "work_request_id": wr.id,
                },
                artifact_filters={"category": ArtifactCategory.BLHC},
            )
            wr.add_event_reaction("on_success", action)
            wr.add_event_reaction("on_failure", action)

    def _populate_regression_analysis_callback(
        self, architectures: set[str]
    ) -> None:
        """Create a callback to collate regression analysis results."""
        if self.data.enable_regression_tracking:
            # Checked by a model validator.
            assert self.data.reference_prefix

            regression_analysis, _ = self.work_request_ensure_child_internal(
                task_name="workflow",
                workflow_data=WorkRequestWorkflowData(
                    allow_dependency_failures=True,
                    step="regression-analysis",
                    display_name="Regression analysis",
                    visible=False,
                ),
                unblock_strategy=WorkRequest.UnblockStrategy.INCREMENTAL,
            )
            for architecture in sorted(architectures):
                for prefix in (self.data.reference_prefix, self.data.prefix):
                    with suppress(KeyError):
                        self.requires_artifact(
                            regression_analysis,
                            f"internal@collections/name:"
                            f"{prefix}blhc-{architecture}",
                        )
            self.orchestrate_child(regression_analysis)

    def _extract_artifact_details(
        self, artifact: Artifact | None
    ) -> tuple[str | None, int | None, WorkRequest | None]:
        """Extract details from a ``debian:blhc`` artifact."""
        source_version: str | None
        artifact_id: int | None
        wr: WorkRequest | None
        if artifact is None:
            source_version = None
            artifact_id = None
            wr = None
        else:
            data = artifact.create_data()
            assert isinstance(data, DebianBlhc)
            source_version = data.version
            artifact_id = artifact.id
            wr = artifact.created_by_work_request
        return source_version, artifact_id, wr

    def callback_regression_analysis(self) -> bool:
        """
        Analyze regressions compared to reference results.

        This is called once for each architecture, but updates the whole
        analysis for all architectures each time.  This is partly for
        simplicity and robustness (we don't need to work out how to combine
        the new analysis with a previous one), and partly to make it easier
        to handle cases where there isn't a one-to-one mapping between the
        reference results and the new results.
        """
        # Select the newest result for each architecture.
        reference_artifacts = self.find_reference_artifacts("blhc")
        new_artifacts = self.find_new_artifacts("blhc", ArtifactCategory.BLHC)

        output_data = self.work_request.output_data or OutputData()
        output_data.regression_analysis = {}
        for architecture in sorted(
            set(reference_artifacts) | set(new_artifacts)
        ):
            reference = reference_artifacts.get(architecture)
            reference_source_version, reference_artifact_id, reference_wr = (
                self._extract_artifact_details(reference)
            )
            new = new_artifacts.get(architecture)
            new_source_version, new_artifact_id, new_wr = (
                self._extract_artifact_details(new)
            )

            output_data.regression_analysis[architecture] = RegressionAnalysis(
                original_source_version=reference_source_version,
                original_artifact_id=reference_artifact_id,
                new_source_version=new_source_version,
                new_artifact_id=new_artifact_id,
                status=self.compare_qa_results(reference_wr, new_wr),
            )

        self.work_request.output_data = output_data
        self.work_request.save()
        return True

    @override
    def get_callback(self, step: str) -> Callable[[], bool]:
        # Current versions of this workflow create a single incremental
        # callback for all architectures called "regression-analysis", but
        # we may be run on workflows that were orchestrated when we created
        # one "regression-analysis-{architecture}" callback per
        # architecture.
        if not step.startswith("regression-analysis"):
            raise NotImplementedError(
                f"Unhandled workflow callback step: {step}"
            )
        return self.callback_regression_analysis

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return the list of input artifacts."""
        artifacts: list[InputArtifact] = []

        if self.data.source_artifact is not None:
            artifacts.append(
                InputArtifactSingle(
                    lookup=self.data.source_artifact,
                    label="source_artifact",
                    artifact_id=(
                        self.dynamic_data.source_artifact_id
                        if self.dynamic_data is not None
                        else None
                    ),
                )
            )

        if self.data.binary_artifacts is not None:
            artifacts.append(
                InputArtifactMultiple.from_ids(
                    lookup=self.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=(
                        self.dynamic_data.binary_artifacts_ids
                        if self.dynamic_data is not None
                        else None
                    ),
                )
            )

        if self.data.package_build_logs is not None:
            artifacts.append(
                InputArtifactMultiple.from_ids(
                    lookup=self.data.package_build_logs,
                    label="package_build_logs",
                    artifact_ids=(
                        self.dynamic_data.package_build_logs_ids
                        if self.dynamic_data is not None
                        else None
                    ),
                )
            )

        return artifacts



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Wait task to ask a user for confirmation."""

from typing import TYPE_CHECKING, assert_never, override

from debusine.server.tasks.wait import BaseWaitTask
from debusine.server.tasks.wait.models import ConfirmData, DenyAction
from debusine.tasks.models import Confirmation, OutputData

if TYPE_CHECKING:
    from debusine.db.models import User


class Confirm(BaseWaitTask[ConfirmData]):
    """Task that waits for user confirmation."""

    TASK_VERSION = 1

    @override
    def get_label(self) -> str:
        return "wait for confirmation"

    def respond(
        self, *, user: "User", confirm: bool, comment: str | None = None
    ) -> None:
        """
        Process a response from the user.

        This is called on the server, either from the API or the web UI.
        """
        from debusine.db.models import WorkRequest

        output_data = OutputData(
            confirmation=Confirmation(confirmed=confirm, comment=comment)
        )
        if confirm:
            self.work_request.mark_completed(
                WorkRequest.Results.SUCCESS, user=user, output_data=output_data
            )
        else:
            match self.data.deny_action:
                case DenyAction.FAIL:
                    self.work_request.mark_completed(
                        WorkRequest.Results.FAILURE,
                        user=user,
                        output_data=output_data,
                    )
                case DenyAction.ABORT:
                    self.work_request.abort(output_data=output_data)
                case DenyAction.ABORT_WORKFLOW:
                    # If this isn't part of a workflow, just abort the work
                    # request itself.  This is easier than trying to
                    # validate that this deny_action can only be used in
                    # work requests that are part of a workflow, and it
                    # should be close enough to what the user intended.
                    (
                        self.work_request.workflow_root or self.work_request
                    ).abort(output_data=output_data)
                case _ as unreachable:
                    assert_never(unreachable)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the system_information."""

import subprocess
from typing import override
from unittest import mock
from unittest.mock import MagicMock

import psutil

import debusine.worker.tags as wtags
from debusine.tasks.models import WorkerType
from debusine.test import TestCase
from debusine.worker import system_information


class TestSystemInformation(TestCase):
    """Test for system_information.py functions."""

    @override
    def setUp(self) -> None:
        """Ensure that native_architecture() is not cached."""
        super().setUp()
        system_information.native_architecture.cache_clear()
        self.addCleanup(system_information.native_architecture.cache_clear)
        self.enterContext(
            mock.patch(
                "debusine.worker.system_information.debusine_version",
                return_value="0.1.0",
            )
        )

    def test_total_physical_memory(self) -> None:
        """Test total_physical_memory function."""
        total_physical_memory = system_information.total_physical_memory()

        # Assert that total_physical_memory "looks good": it is an integer
        # and at least 128 MB of memory
        self.assertIsInstance(total_physical_memory, int)
        self.assertGreaterEqual(total_physical_memory, 128 * 1024 * 1024)

        # Check that it matches the underlying implementation
        self.assertEqual(total_physical_memory, psutil.virtual_memory().total)

    def test_total_cpu_count(self) -> None:
        """Test cpu_count function."""
        cpu_count = system_information.cpu_count()

        # Assert that cpu_count "looks good": it is an integer
        # and at least 1 CPU
        self.assertIsInstance(cpu_count, int)
        self.assertGreaterEqual(cpu_count, 1)

        # Check that it matches the underlying implementation
        self.assertEqual(cpu_count, psutil.cpu_count())

    def test_total_cpu_count_undefined(self) -> None:
        """Test cpu_count function when undefined."""
        with mock.patch("psutil.cpu_count", return_value=None):
            self.assertEqual(system_information.cpu_count(), 1)

    def test_native_architecture(self) -> None:
        """Test native_architecture() function."""
        architecture = system_information.native_architecture()

        # Assert that architecture "looks good": type, length, no end of line
        assert isinstance(architecture, str)
        self.assertGreaterEqual(len(architecture), 1)
        self.assertEqual(architecture, architecture.strip())

        self.assertEqual(
            architecture,
            subprocess.check_output(
                ["dpkg", "--print-architecture"], text=True
            ).strip(),
        )

    def patch_subprocess_check_output(self) -> MagicMock:
        """Patch subprocess.check_output. Return its mock."""
        patcher = mock.patch("subprocess.check_output", autospec=True)
        mocked = patcher.start()
        self.addCleanup(patcher.stop)
        return mocked

    def test_native_architecture_dpkg_not_available(self) -> None:
        """
        Test native_architecture function: return None.

        dpkg is not available or returns an error.
        """
        mocked = self.patch_subprocess_check_output()
        mocked.side_effect = FileNotFoundError()
        self.assertIsNone(system_information.native_architecture())

        system_information.native_architecture.cache_clear()
        mocked.side_effect = subprocess.CalledProcessError(
            1, ["dpkg", "--print-architecture"]
        )
        self.assertIsNone(system_information.native_architecture())

    def test_kernel_supported_architectures_arch_test_fails(self) -> None:
        """
        Test kernel_supported_architectures return None.

        arch-test returns an error.
        """
        mocked = self.patch_subprocess_check_output()

        mocked.side_effect = subprocess.CalledProcessError(
            1, ["arch-test", "-n"]
        )
        self.assertIsNone(system_information.kernel_supported_architectures())

    def test_kernel_supported_architectures(self) -> None:
        """Test kernel_supported_architectures returns a list of archs."""
        mocked = self.patch_subprocess_check_output()

        architectures = ["x1", "x2"]
        mocked.side_effect = ["x2\nx1\n"]

        self.assertEqual(
            system_information.kernel_supported_architectures(), architectures
        )

    def test_system_metadata_without_native_kernel_supported_architectures(
        self,
    ) -> None:
        """
        system_metadata() does not include architectures related keys.

        Functions native_architecture and kernel_supported_architectures
        returned None.
        """
        with (
            mock.patch(
                "debusine.worker.system_information.native_architecture",
                return_value=None,
            ) as architecture_mocked,
            mock.patch(
                "debusine.worker.system_information"
                ".kernel_supported_architectures",
                return_value=None,
            ) as kernel_supported_architectures_mocked,
        ):
            metadata = system_information.system_metadata(WorkerType.EXTERNAL)

        architecture_mocked.assert_called()
        kernel_supported_architectures_mocked.assert_called()

        self.assertNotIn("system:native_architecture", metadata)
        self.assertNotIn("system:architectures", metadata)

    def test_system_metadata_architectures_same_as_native_architecture(
        self,
    ) -> None:
        """
        system_metadata() returns valid metadata for an external worker.

        Method system_information.kernel_supported_architectures() returns
        None: architectures is the same as native_architecture.
        """
        total_physical_memory = system_information.total_physical_memory()
        cpu_count = system_information.cpu_count()
        native_architecture = system_information.native_architecture()

        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=None,
        ):
            self.assertEqual(
                system_information.system_metadata(WorkerType.EXTERNAL),
                {
                    "system:cpu_count": cpu_count,
                    "system:total_physical_memory": total_physical_memory,
                    "system:worker_type": WorkerType.EXTERNAL,
                    "system:native_architecture": native_architecture,
                    "system:architectures": [native_architecture],
                    "system:worker_version": "0.1.0",
                },
            )

    def test_system_metadata_architectures_different_to_native_architecture(
        self,
    ) -> None:
        """system_metadata() populates architectures."""
        native_architecture = system_information.native_architecture()
        kernel_supported_architectures = ["x1", "x2"]
        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=kernel_supported_architectures,
        ):
            self.assertDictContainsAll(
                system_information.system_metadata(WorkerType.EXTERNAL),
                {
                    "system:native_architecture": native_architecture,
                    "system:architectures": kernel_supported_architectures,
                },
            )

    def test_system_provided_tags_without_native_kernel_supported_architectures(
        self,
    ) -> None:
        """
        system_provided_tags() does not include build-arch tags.

        Functions native_architecture and kernel_supported_architectures
        returned None.
        """
        with (
            mock.patch(
                "debusine.worker.system_information.native_architecture",
                return_value=None,
            ) as architecture_mocked,
            mock.patch(
                "debusine.worker.system_information"
                ".kernel_supported_architectures",
                return_value=None,
            ) as kernel_supported_architectures_mocked,
        ):
            tags = system_information.system_provided_tags(WorkerType.EXTERNAL)

        architecture_mocked.assert_called()
        kernel_supported_architectures_mocked.assert_called()

        self.assertEqual(
            [t for t in tags if t.startswith(wtags.BUILD_ARCH_PREFIX)], []
        )

    def test_system_provided_tags_arch_same_as_native(
        self,
    ) -> None:
        """
        system_provided_tags() returns native architecture.

        Method system_information.kernel_supported_architectures() returns
        None: architectures is the same as native_architecture.
        """
        native_architecture = system_information.native_architecture()
        assert native_architecture is not None

        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=None,
        ):
            tags = system_information.system_provided_tags(WorkerType.EXTERNAL)

        self.assertEqual(
            [t for t in tags if t.startswith(wtags.BUILD_ARCH_PREFIX)],
            [wtags.BUILD_ARCH_PREFIX + native_architecture],
        )

    def test_system_provided_tags_arch_different_than_native(
        self,
    ) -> None:
        """system_provided_tags() returns all available architectures."""
        native_architecture = system_information.native_architecture()
        assert native_architecture is not None
        kernel_supported_architectures = ["x1", "x2"]
        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=kernel_supported_architectures,
        ):
            tags = system_information.system_provided_tags(WorkerType.EXTERNAL)

        self.assertCountEqual(
            [t for t in tags if t.startswith(wtags.BUILD_ARCH_PREFIX)],
            [
                wtags.BUILD_ARCH_PREFIX + native_architecture,
                wtags.BUILD_ARCH_PREFIX + "x1",
                wtags.BUILD_ARCH_PREFIX + "x2",
            ],
        )

    def test_system_metadata_celery(self) -> None:
        """system_metadata() returns valid metadata for a Celery worker."""
        total_physical_memory = system_information.total_physical_memory()
        cpu_count = system_information.cpu_count()
        native_architecture = system_information.native_architecture()

        system_metadata = system_information.system_metadata(WorkerType.CELERY)

        # system:architectures is tested in other unit tests
        system_metadata.pop("system:architectures")

        self.assertEqual(
            system_metadata,
            {
                "system:cpu_count": cpu_count,
                "system:total_physical_memory": total_physical_memory,
                "system:worker_type": WorkerType.CELERY,
                "system:native_architecture": native_architecture,
                "system:worker_version": "0.1.0",
            },
        )

    def test_system_metadata_signing(self) -> None:
        """system_metadata() returns valid metadata for a signing worker."""
        total_physical_memory = system_information.total_physical_memory()
        cpu_count = system_information.cpu_count()
        native_architecture = system_information.native_architecture()

        system_metadata = system_information.system_metadata(WorkerType.SIGNING)

        # system:architectures is tested in other unit tests
        system_metadata.pop("system:architectures")

        self.assertEqual(
            system_metadata,
            {
                "system:cpu_count": cpu_count,
                "system:total_physical_memory": total_physical_memory,
                "system:worker_type": WorkerType.SIGNING,
                "system:native_architecture": native_architecture,
                "system:worker_version": "0.1.0",
            },
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Context information needed to check permissions."""

from dataclasses import dataclass, field, replace
from typing import Self, TYPE_CHECKING, overload, override

from django.contrib.auth.models import AnonymousUser
from django.db.models import Q, QuerySet

if TYPE_CHECKING:
    from debusine.db.models import (
        CollectionTokenGrant,
        Group,
        Token,
        User,
        WorkRequest,
    )
    from debusine.db.models.auth import GroupQuerySet

#: Type alias for the user variable used by permission predicates
type PermissionUser = "User | AnonymousUser"


@dataclass(frozen=True)
class PermissionContext:
    """Context information passed to permission predicates."""

    @staticmethod
    def _extra_groups_default_factory() -> "GroupQuerySet[Group]":
        from debusine.db.models import Group

        return Group.objects.none()

    @staticmethod
    def _extra_collection_grants_default_factory() -> (
        "QuerySet[CollectionTokenGrant]"
    ):
        from debusine.db.models import CollectionTokenGrant

        return CollectionTokenGrant.objects.none()

    #: The current user.
    user: PermissionUser
    #: Extra groups used for additional permissions.
    extra_groups: "GroupQuerySet[Group]" = field(
        default_factory=_extra_groups_default_factory,
        # Ignore extra_groups for comparisons, since it is derived from
        # work_request.
        compare=False,
    )
    #: Extra roles on resources granted to specific authorization contexts.
    extra_collection_grants: "QuerySet[CollectionTokenGrant]" = field(
        default_factory=_extra_collection_grants_default_factory,
        # Ignore extra_collection_grants for comparisons, since it is
        # derived from the token.
        compare=False,
    )
    #: True if permission checks should always succeed.
    #: This is used to skip permission checks in situations like management
    #: commands or test fixture setup.
    permission_checks_disabled: bool = False
    #: The current user token, if used.
    user_token: "Token | None" = None
    #: The current external user token, if used.
    external_user_token: "Token | None" = None
    #: The current worker token, if used.
    worker_token: "Token | None" = None
    #: The current work request token, if used.
    work_request_token: "Token | None" = None
    #: The current worker activation token, if used
    worker_activation_token: "Token | None" = None
    #: The current work request, if used.
    work_request: "WorkRequest | None" = None

    @override
    def __str__(self) -> str:
        res = str(self.user)
        if self.user_token:
            res += "[user token]"
        if self.external_user_token:
            res += "[external user token]"
        if self.worker_token:
            res += "[worker token]"
        if self.worker_activation_token:
            res += "[worker-activation token]"
        if self.work_request_token:
            res += "[work-request token]"
        if self.work_request:
            res += f"[work request {self.work_request}]"
        extra_groups = tuple(self.extra_groups.order_by("pk"))
        if extra_groups:
            res += f"[extra groups {', '.join(str(g) for g in extra_groups)}]"
        return res

    @override
    def __repr__(self) -> str:
        return str(self)

    def with_permission_checks_disabled(self) -> Self:
        """Return a copy of this PermissionContext with checks disabled."""
        return replace(self, permission_checks_disabled=True)

    def groups(self) -> "GroupQuerySet[Group]":
        """Return a queryset selecting the active groups."""
        # Prevent circular import
        from debusine.db.models import Group

        if not self.user.is_authenticated:
            return Group.objects.none()

        return Group.objects.filter(
            Q(users=self.user) | Q(pk__in=self.extra_groups.values("pk"))
        )

    @property
    def explicit_grants_only(self) -> bool:
        """Return whether this context's token requires explicit grants."""
        token = (
            self.user_token
            or self.external_user_token
            or self.work_request_token
        )
        if token is not None:
            return token.explicit_grants_only
        else:
            return False

    @property
    def is_non_permission_bearing_token(self) -> bool:
        """Return whether the token is one that cannot grant any permissions."""
        return (self.worker_token or self.worker_activation_token) is not None

    @classmethod
    def for_ui_anonymous(cls) -> Self:
        """Return a PermissionContext for an anonymous user in the UI."""
        return cls(user=AnonymousUser())

    @classmethod
    def for_ui_user(cls, user: "User") -> Self:
        """Return a PermissionContext for a user authenticating in the UI."""
        return cls(user=user)

    @classmethod
    def for_api_user(cls, token: "Token") -> Self:
        """Return a PermissionContext for a user authenticating in the API."""
        # Prevent circular import
        from debusine.db.models import Token

        assert token.token_type == Token.Types.USER
        assert token.enabled
        assert token.user is not None
        return cls(user_token=token, user=token.user)

    @classmethod
    def for_api_external_user(cls, token: "Token") -> Self:
        """
        Return a PermissionContext for an external user via the API.

        External users have a username, but do not correspond to Debusine
        users.  They largely act as the anonymous user, but with specific
        additional grants.
        """
        from debusine.db.models import Token

        assert token.token_type == Token.Types.EXTERNAL_USER
        assert token.enabled
        return cls(
            user=AnonymousUser(),
            external_user_token=token,
            extra_collection_grants=token.collection_grants.all(),
        )

    @classmethod
    def for_worker(cls, token: "Token") -> Self:
        """Return a PermissionContext for a worker connecting to Debusine."""
        from debusine.db.models import Token

        assert token.token_type == Token.Types.WORKER
        assert token.enabled
        assert token.user is None  # enforced by database constraint
        return cls(user=AnonymousUser(), worker_token=token)

    @overload
    @classmethod
    def for_work_request(cls, *, token: "Token") -> Self: ...

    @overload
    @classmethod
    def for_work_request(cls, *, work_request: "WorkRequest") -> Self: ...

    @classmethod
    def for_work_request(
        cls,
        *,
        token: "Token | None" = None,
        work_request: "WorkRequest | None" = None,
    ) -> Self:
        """Return a PermissionContext for a worker processing a work request."""
        from debusine.db.models import Token

        if token is not None:
            assert work_request is None
            assert token.token_type == Token.Types.WORK_REQUEST
            assert token.work_request is not None
            assert token.enabled
            return cls(
                user=token.work_request.created_by,
                extra_groups=token.work_request.get_extra_groups(),
                work_request_token=token,
                work_request=token.work_request,
            )
        elif work_request is not None:
            return cls(
                user=work_request.created_by,
                extra_groups=work_request.get_extra_groups(),
                work_request=work_request,
            )
        else:
            raise AssertionError("neither token nor work_request passed")

    @classmethod
    def for_worker_activation(cls, token: "Token") -> Self:
        """Return a PermissionContext for worker activation."""
        from debusine.db.models import Token

        assert token.token_type == Token.Types.WORKER_ACTIVATION
        assert token.enabled
        return cls(user=AnonymousUser(), worker_activation_token=token)

    @classmethod
    def for_maintenance(cls) -> Self:
        """Return a PermissionContext for maintenance commands."""
        from debusine.db.models.auth import system_user

        return cls(user=system_user())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used by server tasks."""

from typing import Any

from debusine.db.tasks.models import (
    DBInputCollectionSingle,
    DBInputLookupResult,
)
from debusine.tasks.models import BaseLabeledTaskInput


class CopyCollectionItemsCopiesInfo(BaseLabeledTaskInput):
    """A resolved set of copies for the CopyCollectionItems task."""

    source_items: list[DBInputLookupResult]
    target_collection: DBInputCollectionSingle
    unembargo: bool
    replace: bool
    name_template: str | None
    variables: dict[str, Any] | None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for package archives."""

import datetime as dt
import logging
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Any, override

from debian import deb822
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.handlers.exception import response_for_exception
from django.db.models import QuerySet
from django.db.models.fields.json import KT
from django.http import Http404, HttpRequest, HttpResponse, HttpResponseBase
from django.shortcuts import get_object_or_404
from django.utils.cache import patch_cache_control, patch_vary_headers
from django.views.generic import TemplateView
from pgpy import PGPKey

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.assets.asset_utils import ExportablePGPKeyring
from debusine.db.context import context
from debusine.db.models import (
    Asset,
    Collection,
    CollectionItem,
    CollectionRelation,
    FileInArtifact,
)
from debusine.db.models.assets import get_public_keys
from debusine.server.collections.debian_suite import (
    DebianSuiteManager,
    make_source_prefix,
)
from debusine.web.archives.converters import SnapshotConverter
from debusine.web.views.base import BaseUIView
from debusine.web.views.files import FileDownloadMixin, FileUI
from debusine.web.views.places import Place


@dataclass(frozen=True, kw_only=True)
class RenderableSuite:
    """Information about a suite that can be rendered via a template."""

    request_scheme: str
    snapshot: dt.datetime | None = None
    collection: Collection

    @property
    def collection_manager(self) -> DebianSuiteManager:
        """The collection manager for the suite."""
        return DebianSuiteManager(self.collection)

    @property
    def name(self) -> str:
        """The name of the suite."""
        return self.collection.name

    @property
    def components(self) -> list[str] | None:
        """The components that exist in this suite."""
        components = self.collection.data.get("components")
        assert isinstance(components, (list, type(None)))
        return components

    @property
    def architectures(self) -> list[str] | None:
        """The architectures that exist in this suite."""
        architectures = self.collection.data.get("architectures")
        assert isinstance(architectures, (list, type(None)))
        return architectures

    @property
    def release_fields(self) -> dict[str, str]:
        """Static fields to set in this suite's ``Release`` file."""
        release_fields = self.collection.data.get("release_fields", {})
        assert isinstance(release_fields, dict)
        return release_fields

    @property
    def description(self) -> str | None:
        """The description of the suite, if any."""
        return self.release_fields.get("Description")

    def as_deb822_source(self) -> str:
        """Render a deb822-style APT source."""
        archive_fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN
        workspace = self.collection.workspace
        source = deb822.Deb822()
        source["Types"] = "deb"
        source["URIs"] = (
            f"{self.request_scheme}://{archive_fqdn}"
            f"/{workspace.scope.name}/{workspace.name}"
        )
        source["Suites"] = self.collection.name
        if self.components is not None:
            source["Components"] = " ".join(self.components)
        if signing_keys := self.collection_manager.export_signing_keys():
            source["Signed-By"] = "\n" + "\n".join(
                " " + (line or ".") for line in signing_keys.splitlines()
            )
        if self.snapshot is not None:
            source["Snapshot"] = SnapshotConverter().to_url(self.snapshot)
        return source.dump()

    def get_required_renderable_suites(self) -> list["RenderableSuite"]:
        """Return renderable information for suites recursively required."""
        suites: list[RenderableSuite] = []

        seen_collection_ids = {self.collection.id}

        def add_required_suites(collection: Collection) -> None:
            relations = (
                collection.relations.filter(
                    type=CollectionRelation.Types.SUITE_REQUIRES,
                    target__category=CollectionCategory.SUITE,
                )
                .select_related("target__workspace__scope")
                .order_by("position")
            )

            for relation in relations:
                target = relation.target
                if target.id in seen_collection_ids:
                    continue

                seen_collection_ids.add(target.id)

                suites.append(
                    RenderableSuite(
                        request_scheme=self.request_scheme,
                        snapshot=self.snapshot,
                        collection=target,
                    )
                )

                add_required_suites(target)

        add_required_suites(self.collection)
        return suites

    def required_suites_as_deb822_sources(self) -> str:
        """Render deb822-style APT sources required by this suite."""
        return "\n".join(
            suite.as_deb822_source()
            for suite in self.get_required_renderable_suites()
        )


class SigningKeysResponse(HttpResponse):
    """A renderer for OpenPGP signing keys."""

    def __init__(self, keys: list[bytes]) -> None:
        """Construct an HTTP response from a list of keys."""
        content = str(
            ExportablePGPKeyring(*(PGPKey.from_blob(key)[0] for key in keys))
        )
        super().__init__(
            content=content.encode("utf-8"),
            content_type="application/pgp-keys; charset=utf-8",
        )


class ArchiveView(BaseUIView, ABC):
    """A view of something in an archive."""

    base_template = "web/_base_archives.html"
    archive: Collection

    @property
    def cache_control(self) -> dict[str, Any]:
        """Return appropriate ``Cache-Control`` arguments for this response."""
        # Responses resulting from a snapshot-specific query are immutable.
        # For other responses, default to telling caches to consider them
        # stale after a relatively short delay; subclasses may override
        # this.
        return (
            {"max-age": 31536000}
            if self.kwargs.get("snapshot") is not None
            else {"max-age": 1800, "proxy-revalidate": True}
        )

    def get_snapshot(self) -> QuerySet[CollectionItem]:
        """Return a queryset of items active at the selected timestamp."""
        qs = CollectionItem.objects.filter(
            parent_collection__workspace=context.require_workspace()
        )
        if (snapshot := self.kwargs.get("snapshot")) is not None:
            return qs.active_at(snapshot)
        else:
            return qs.active()

    def get_suites(self) -> QuerySet[Collection]:
        """Return a queryset of relevant suites."""
        # Collections don't have created/removed times, but that isn't a big
        # problem for snapshot support since the items they contain do.
        qs = (
            Collection.objects.in_current_workspace()
            .exported_suites()
            .can_display(context.pc)
        )
        if (suite := self.kwargs.get("suite")) is not None:
            qs = qs.filter(name=suite)
        return qs

    @cached_property
    def suite(self) -> Collection | None:
        """The current suite, if this is a view for a specific suite."""
        if "suite" in self.kwargs:
            try:
                return self.get_suites().get()
            except Collection.DoesNotExist as exc:
                if (workspace := context.require_workspace()).public:
                    raise Http404("Suite not found") from exc
                else:
                    # Report this as if the workspace as a whole did not
                    # exist, to avoid information leaks.
                    logging.debug(
                        "permission denied on suite in %s reported as 404",
                        workspace,
                    )
                    self.raise_workspace_not_found(workspace)
        else:
            return None

    def enforce_visibility(self) -> None:
        """Enforce visibility of the archive or the suite, as appropriate."""
        collection = self.suite or self.archive
        try:
            self.enforce(collection.can_display)
        except PermissionDenied:
            # Report this as if the workspace as a whole did not exist, to
            # avoid information leaks.
            logging.debug("permission denied on %s reported as 404", collection)
            self.raise_workspace_not_found(collection.workspace)

    def set_archive(self) -> None:
        """Set the archive based on the request arguments."""
        self.set_current_workspace(
            self.kwargs["workspace"], check_visibility=False
        )
        workspace = context.require_workspace()
        try:
            self.archive = Collection.objects.get(
                workspace=workspace, category=CollectionCategory.ARCHIVE
            )
        except Collection.DoesNotExist:
            raise Http404(
                f"No {CollectionCategory.ARCHIVE} collection found in "
                f"{workspace}"
            )
        self.enforce_visibility()

    @override
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """Dispatch the request, setting appropriate response headers."""
        try:
            response = super().dispatch(request, *args, **kwargs)
        except Exception as exc:
            response = response_for_exception(request, exc)
        patch_cache_control(response, **self.cache_control)
        patch_vary_headers(response, ("Authorization",))
        return response


class ArchiveRootView(ArchiveView, TemplateView):
    """The root of an archive."""

    template_name = "web/archive-root.html"

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        self.set_archive()
        ctx = super().get_context_data(**kwargs)

        suites = [
            RenderableSuite(
                request_scheme=self.request.scheme or "http",
                snapshot=self.kwargs.get("snapshot"),
                collection=suite,
            )
            for suite in self.get_suites().order_by("name")
        ]

        ctx["suites"] = suites
        return ctx

    @override
    def get_place(self) -> Place:
        assert context.workspace is not None
        return context.workspace.ui(self.request).place_suite_list


class SuiteRootView(ArchiveView, TemplateView):
    """The root of a suite."""

    template_name = "web/suite-root.html"

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        self.set_archive()

        ctx = super().get_context_data(**kwargs)

        assert self.suite is not None
        ctx["suite"] = RenderableSuite(
            request_scheme=self.request.scheme or "http",
            snapshot=self.kwargs.get("snapshot"),
            collection=self.suite,
        )
        return ctx

    @override
    def get_place(self) -> Place:
        assert self.suite is not None
        return self.suite.ui(self.request).place_suite_root


class SigningKeysView(ArchiveView):
    """Signing keys for an archive or a suite."""

    def get(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        *args: Any,
        **kwargs: Any,
    ) -> HttpResponse:
        """Show signing keys for an archive or a suite."""
        self.set_archive()
        signing_keys: set[Asset] = set()
        for suite in self.get_suites():
            signing_keys.update(
                DebianSuiteManager(suite).get_signing_keys() or []
            )
        return SigningKeysResponse(list(get_public_keys(signing_keys)))


class ArchiveFileView(ArchiveView, FileDownloadMixin, ABC):
    """A file in an archive."""

    @abstractmethod
    def get_collection_items(self) -> QuerySet[CollectionItem]:
        """
        Return a queryset of collection items matching the requested file.

        Subclasses should use their URL arguments to narrow the filter until
        it is guaranteed to return at most one item pointing to an artifact
        containing exactly one file.
        """
        return self.get_snapshot().filter(
            child_type=CollectionItem.Types.ARTIFACT, artifact__isnull=False
        )

    def get_queryset(self) -> QuerySet[FileInArtifact]:
        """Filter to a single matching file."""
        # Maybe this should use the generic collection lookup mechanism, but
        # that doesn't currently support snapshots or other things we need
        # such as constraining pool files by component and source package
        # name.  In any case, it seems worth having an optimized lookup path
        # for these views.
        return FileInArtifact.objects.filter(
            artifact__in=self.get_collection_items().values("artifact")
        )

    def get(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        *args: Any,
        **kwargs: Any,
    ) -> HttpResponseBase:
        """Download a file."""
        self.set_archive()
        try:
            file_in_artifact = get_object_or_404(self.get_queryset())
        except Http404:
            if (workspace := context.require_workspace()).public:
                raise
            else:
                # Report this as if the workspace as a whole did not exist,
                # to avoid information leaks.
                logging.debug(
                    "permission denied on file in %s reported as 404", workspace
                )
                self.raise_workspace_not_found(workspace)
        ui_info = FileUI.from_file_in_artifact(file_in_artifact)
        return self.stream_file(file_in_artifact, ui_info)


class SuiteFileView(ArchiveFileView, ABC):
    """A file looked up via a suite."""

    @override
    @abstractmethod
    def get_collection_items(self) -> QuerySet[CollectionItem]:
        """Filter to items in any suite in the selected archive."""
        return (
            super()
            .get_collection_items()
            .filter(
                parent_collection__in=self.get_suites(),
                # Technically redundant with the filters in
                # self.get_suites(), but helps PostgreSQL to use the correct
                # index.
                parent_category=CollectionCategory.SUITE,
            )
        )

    @override
    def enforce_visibility(self) -> None:
        # A narrowly-scoped token might only have permission to see an
        # individual suite within the archive, rather than the whole
        # archive.  Such tokens should be able to see any file referenced
        # via any suite that they can see.  Since this view filters to items
        # in any of the suites returned by get_suites, and since get_suites
        # only returns visible suites, we can safely skip further
        # permissions enforcement.
        pass


class DistsByHashFileView(SuiteFileView):
    """An index file in a ``by-hash`` directory."""

    @override
    @property
    def cache_control(self) -> dict[str, Any]:
        """``by-hash`` responses are always immutable."""
        return {"max-age": 31536000}

    @override
    def get_snapshot(self) -> QuerySet[CollectionItem]:
        """
        Return a queryset of items that existed at the selected timestamp.

        This is slightly different for ``by-hash`` files than others: the
        items need to have existed at the given timestamp, but they don't
        need to have been active.
        """
        qs = CollectionItem.objects.filter(
            parent_collection__workspace=context.require_workspace()
        )
        if (snapshot := self.kwargs.get("snapshot")) is not None:
            qs = qs.filter(created_at__lte=snapshot)
        return qs

    @override
    def get_collection_items(self) -> QuerySet[CollectionItem]:
        """Filter to the requested index file."""
        return (
            super()
            .get_collection_items()
            .annotate(path=KT("data__path"))
            .filter(
                category=ArtifactCategory.REPOSITORY_INDEX,
                # The directory name must be equal to everything up to the
                # last slash.
                path__regex=f"^{re.escape(self.kwargs['directory'])}/[^/]+$",
            )
            # Release files and friends don't have by-hash entries.
            .exclude(path__in=("Release", "Release.gpg", "InRelease"))
        )

    @override
    def get_queryset(self) -> QuerySet[FileInArtifact]:
        """Filter to a file with the requested checksum."""
        # XXX 404 if bytes.fromhex fails
        # We may match multiple identical files, take the first.
        return (
            super()
            .get_queryset()
            .filter(file__sha256=bytes.fromhex(self.kwargs["checksum"]))
        )[:1]


class DistsFileView(SuiteFileView):
    """An index file in a suite."""

    @override
    def get_collection_items(self) -> QuerySet[CollectionItem]:
        """Filter to the requested index file."""
        return (
            super()
            .get_collection_items()
            .annotate(path=KT("data__path"))
            .filter(
                category=ArtifactCategory.REPOSITORY_INDEX,
                path=self.kwargs["path"],
            )
        )


class PoolFileView(SuiteFileView):
    """
    A package in the archive's pool.

    While the suite isn't in the URL, we still need to look up pool files
    via any suite in the archive.  It may be in more than one suite, but
    constraints guarantee that they all refer to the same contents.
    """

    _re_binary_package = re.compile(
        r"^([^_]+)_([^_]+)_([^.]+)\.(?:deb|ddeb|udeb)$"
    )

    @override
    @property
    def cache_control(self) -> dict[str, Any]:
        """Pool files in ``may_reuse_versions=False`` archives are immutable."""
        if hasattr(self, "archive") and not self.archive.data.get(
            "may_reuse_versions", False
        ):
            return {"max-age": 31536000}
        else:
            return super().cache_control

    @override
    def get_collection_items(self) -> QuerySet[CollectionItem]:
        """Filter to the requested pool file."""
        if self.kwargs["sourceprefix"] != make_source_prefix(
            self.kwargs["source"]
        ):
            raise Http404("No FileInArtifact matches the given query.")

        qs = (
            super()
            .get_collection_items()
            .annotate(component=KT("data__component"))
            .filter(component=self.kwargs["component"])
        )
        # Although filtering on collection item properties here is somewhat
        # redundant with the path filter that get_queryset() will add, it
        # allows us to make use of db_ci_suite_source_idx and
        # db_ci_suite_binary_source_idx, and it means the database can
        # filter out irrelevant CollectionItem rows before needing to look
        # at FileInArtifact.  We don't bother constraining these searches by
        # version, since some files in source packages only contain the
        # upstream part of the version; this also saves us from worrying
        # about epochs.
        if m := self._re_binary_package.match(self.kwargs["filename"]):
            qs = qs.annotate(
                srcpkg_name=KT("data__srcpkg_name"),
                package=KT("data__package"),
                architecture=KT("data__architecture"),
            ).filter(
                category=ArtifactCategory.BINARY_PACKAGE,
                srcpkg_name=self.kwargs["source"],
                package=m.group(1),
                architecture=m.group(3),
            )
        else:
            qs = qs.annotate(package=KT("data__package")).filter(
                category=ArtifactCategory.SOURCE_PACKAGE,
                package=self.kwargs["source"],
            )
        return qs

    @override
    def get_queryset(self) -> QuerySet[FileInArtifact]:
        """Filter to a file with the requested name."""
        return (
            super()
            .get_queryset()
            .filter(path=self.kwargs["filename"])
            .distinct("file")
        )


class TopLevelFileView(ArchiveFileView):
    """A top-level file in the archive, not in any suite."""

    @override
    def get_collection_items(self) -> QuerySet[CollectionItem]:
        """Filter to the requested index file."""
        return (
            super()
            .get_collection_items()
            .annotate(path=KT("data__path"))
            .filter(
                parent_category=CollectionCategory.ARCHIVE,
                category=ArtifactCategory.REPOSITORY_INDEX,
                path=self.kwargs["path"],
            )
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: workers."""

import logging
from typing import Any

import pgtransaction
from celery import shared_task
from django.db import OperationalError
from django.db.models import Exists, OuterRef
from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.db.context import context
from debusine.db.models import Token, WorkRequest, Worker
from debusine.db.models.permissions import anonymous_access
from debusine.db.models.work_requests import (
    CannotRetry,
    MAX_AUTOMATIC_RETRIES,
    WorkRequestRetryReason,
)
from debusine.db.transactions import atomic
from debusine.project.celery import (
    apply_async_on_commit,
    celery_tasks_in_progress,
)
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    WorkRequestCompletedSerializer,
    WorkerRegisterSerializer,
    WorkerWorkRequestSerializer,
)
from debusine.server.views.base import BaseAPIView
from debusine.server.views.rest import (
    IsWorkRequestTokenAuthenticated,
    IsWorkerTokenAuthenticated,
)
from debusine.tasks.models import OutputData, OutputDataError

logger = logging.getLogger(__name__)


class RegisterView(BaseAPIView):
    """View used by workers to register to debusine."""

    def post(self, request: Request) -> Response:
        """Worker registers (sends token and fqdn)."""
        self.enforce(anonymous_access)
        worker_register = WorkerRegisterSerializer(data=request.data)

        if not worker_register.is_valid():
            raise DebusineAPIException(
                title="Cannot deserialize worker",
                validation_errors=worker_register.errors,
            )

        valid_activation_token = (
            isinstance(request.auth, Token)
            and request.auth.token_type == Token.Types.WORKER_ACTIVATION
        )

        if (
            request.headers.get("token") is not None
            and not valid_activation_token
        ):
            raise DebusineAPIException(
                title="Invalid worker activation token",
            )

        token_key = worker_register.validated_data['token']
        fqdn = worker_register.validated_data['fqdn']
        worker_type = worker_register.validated_data['worker_type']

        token_hash = Token._generate_hash(token_key)
        token, _ = Token.objects.get_or_create(
            token_type=Token.Types.WORKER, hash=token_hash
        )

        if valid_activation_token:
            assert isinstance(request.auth, Token)
            # Activate an existing worker by registering and automatically
            # enabling a token for it.
            activating_worker = request.auth.activating_worker
            assert activating_worker is not None
            activating_worker.activation_token = None
            activating_worker.token = token
            activating_worker.save()
            token.refresh_from_db()
            token.enable()
            request.auth.delete()
        else:
            Worker.objects.create_with_fqdn(
                fqdn, token=token, worker_type=worker_type
            )

        logger.info('Client registered. Token key: %s', token_key)

        return Response(status=status.HTTP_201_CREATED)


class GetNextWorkRequestView(BaseAPIView):
    """View used by workers to request a task."""

    permission_classes = [IsWorkerTokenAuthenticated]

    def select_work_request(self, worker: Worker) -> WorkRequest | None:
        """Pick our work request from the assigned queue."""
        running_work_requests = WorkRequest.objects.running(worker=worker)

        # If (1) a worker has recently requested that a work request should
        # be marked as completed, but the server hasn't finished dealing
        # with that yet, and (2) a websocket connection was restarted, then
        # we might find ourselves here.  In that case, we don't want to
        # interfere with a work request that's very nearly finished.
        in_progress_task_results = celery_tasks_in_progress(
            "debusine.server.views.workers.update_work_request_as_completed"
        )
        if running_work_requests.filter(
            Exists(
                in_progress_task_results.filter(first_task_arg=OuterRef("pk"))
            )
        ).exists():
            return None

        # Did the worker fail before producing anything?
        # Then it's safe to just re-attempt execution
        for running in running_work_requests.reassignable(worker=worker):
            workflow_data = running.workflow_data
            if workflow_data.retry_count >= MAX_AUTOMATIC_RETRIES:
                logger.info(
                    (
                        "Cannot retry previously-running work request %i: "
                        "automatic retry limit exceeded."
                    ),
                    running.id,
                )
                running.mark_aborted(
                    output_data=OutputData(
                        errors=[
                            OutputDataError(
                                message=(
                                    f"No worker managed to complete work "
                                    f"request {running.id}. Automatic retry "
                                    f"limit exceeded."
                                ),
                                code="work-request-retry-limit-exceeded",
                            )
                        ]
                    )
                )
                continue
            logger.info("Re-attempting work request %i", running.id)
            workflow_data.retry_count += 1
            running.workflow_data = workflow_data
            running.save()
            return running

        pending_work_requests = WorkRequest.objects.pending(worker=worker)

        # Aborting running work requests may change the first pending work
        # request that we select later, but it will still be in one of these
        # workflows.
        try:
            (
                running_work_requests | pending_work_requests[:1]
            ).lock_workflows_for_update(nowait=True)
        except OperationalError:
            logger.debug(
                "Failed to lock work requests assigned to %s", worker.name
            )
            raise DebusineAPIException(
                title="Retry",
                detail="Could not acquire workflow locks; retry later",
                # This is normally a WebDAV response code, but it seems
                # somewhat reasonable here too.
                status_code=status.HTTP_423_LOCKED,
            )

        # This view is normally only called when a worker is idle, but if
        # the worker was restarted then the server may still think it has a
        # running work request.  In that case, retry the work request since
        # it may be in an unclear intermediate state.
        for running in running_work_requests:
            logger.info("Retrying stuck work request %s", running.pk)
            try:
                # Disable permission checks, as retrying a work request is
                # normally not allowed with a worker token
                with context.disable_permission_checks():
                    running.retry(
                        reason=WorkRequestRetryReason.WORKER_FAILED,
                        force_status=True,
                    )
            except CannotRetry as e:
                # retry(force_status=True) starts by aborting the work
                # request; swallowing this exception and committing the
                # transaction anyway has the effect of leaving the work
                # request in the aborted status, which is what we want.
                logger.debug(  # noqa: G200
                    "Cannot retry previously-running work request: %s", e
                )

        return pending_work_requests.first()

    def post(self, request: Request) -> Response:  # noqa: ARG002
        """Return the task to build."""
        # This can't be None at this point, since IsWorkerTokenAuthenticated
        # passed.
        assert context.pc.worker_token
        worker = context.pc.worker_token.worker
        assert worker is not None

        # restricted to workers
        self.enforce(anonymous_access)

        work_request = self.select_work_request(worker)

        status_code: int
        if work_request:
            content = WorkerWorkRequestSerializer(work_request).data
            work_request.mark_running()
            status_code = status.HTTP_200_OK
        else:
            # There is no work request available for the worker
            content = None
            status_code = status.HTTP_204_NO_CONTENT

        return Response(content, status=status_code)


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task  # type: ignore[untyped-decorator]
@context.local()
@context.disable_permission_checks()
def update_work_request_as_completed(
    work_request_id: int,
    result: WorkRequest.Results,
    output_data: dict[str, Any] | None = None,
) -> None:
    """Mark a work request as completed."""
    found_work_request = False

    @atomic(isolation_level=pgtransaction.REPEATABLE_READ)
    def mark_completed() -> None:
        nonlocal found_work_request

        WorkRequest.objects.filter(
            id=work_request_id
        ).lock_workflows_for_update()

        try:
            work_request = WorkRequest.objects.get(pk=work_request_id)
        except WorkRequest.DoesNotExist:
            logger.error("Work request %d does not exist", work_request_id)
            raise
        found_work_request = True

        work_request.mark_completed(
            result,
            output_data=(
                None
                if output_data is None
                else OutputData.model_validate(output_data)
            ),
        )

    @atomic(isolation_level=pgtransaction.REPEATABLE_READ)
    def retry_if_running() -> None:
        if tuple(
            WorkRequest.objects.filter(pk=work_request_id).values_list(
                "status", flat=True
            )
        ) == (WorkRequest.Statuses.RUNNING,):
            work_request = (
                WorkRequest.objects.filter(id=work_request_id)
                .lock_workflows_for_update()
                .get()
            )

            # Since the worker won't be informed of this Celery task
            # failing, we must take care to avoid it becoming stuck.  Retry
            # this so that the scheduler will give the worker something else
            # to do.
            work_request.retry(
                reason=WorkRequestRetryReason.WORKER_FAILED,
                force_status=True,
            )

    try:
        mark_completed()
    except Exception:
        if not found_work_request:
            # Re-raise any exceptions from before we got as far as trying to
            # mark the work request as completed.
            raise

        logger.exception(
            "Failed to mark work request %d as completed", work_request_id
        )
        retry_if_running()


class UpdateWorkRequestAsCompletedView(BaseAPIView):
    """
    View used by the workers to mark a task as completed.

    This defers most of the work to a Celery task, to avoid lock contention
    causing timeouts and retries, and potentially even workers having to
    give up and lose work.

    There is a slight risk in deferring this work.  If the worker is
    restarted between receiving the result from this view and the Celery
    task completing, then the server may send it another
    ``work_request_available`` message, causing the worker to request the
    next work request, causing the server to think that the worker failed
    and retry the work request.  However, given a successful response from
    this view, that scenario should only happen if the worker is restarted
    manually.
    """

    permission_classes = [IsWorkRequestTokenAuthenticated]

    def put(self, request: Request, work_request_id: int) -> Response:
        """Mark a work request as completed."""
        self.enforce(anonymous_access)  # restricted to workers

        try:
            work_request = WorkRequest.objects.can_display(context.pc).get(
                pk=work_request_id
            )
        except WorkRequest.DoesNotExist:
            raise DebusineAPIException(
                title="Work request not found",
                status_code=status.HTTP_404_NOT_FOUND,
            )

        self.check_object_permissions(request, work_request)

        work_request_completed_serializer = WorkRequestCompletedSerializer(
            data=request.data
        )

        if work_request_completed_serializer.is_valid():
            data = work_request_completed_serializer.validated_data
            apply_async_on_commit(
                update_work_request_as_completed,
                args=(work_request_id,),
                kwargs={
                    "result": data["result"],
                    "output_data": data.get("output_data"),
                },
            )
            content = None
            status_code = status.HTTP_200_OK
        else:
            raise DebusineAPIException(
                title="Cannot change work request as completed",
                validation_errors=work_request_completed_serializer.errors,
            )

        return Response(content, status=status_code)


class UpdateWorkerDynamicMetadataView(BaseAPIView):
    """View used by the workers to post dynamic metadata."""

    permission_classes = [IsWorkerTokenAuthenticated]

    def put(self, request: Request) -> Response:
        """Update Worker dynamic metadata."""
        self.enforce(anonymous_access)
        token_key = request.headers['token']
        worker = Worker.objects.get_worker_by_token_key_or_none(token_key)
        # Checked by IsWorkerTokenAuthenticated.
        assert worker is not None

        worker.set_dynamic_metadata(request.data)

        return Response(status=status.HTTP_204_NO_CONTENT)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for Debusine dput-ng hooks."""

import logging
from unittest import mock

from dput.changes import Changes
from dput.exceptions import HookException

from debusine.client.dput_ng.hooks import (
    NoWorkflowError,
    check_workflow,
    choose_workflow,
    create_workflow,
)
from debusine.client.dput_ng.tests.test_utils import (
    FakeInterface,
    load_local_profile,
    mock_client_config,
)
from debusine.client.exceptions import DebusineError
from debusine.client.models import CreateWorkflowRequest
from debusine.test import TestCase
from debusine.test.test_utils import create_work_request_response


class TestChooseWorkflow(TestCase):
    """Test :py:func:`choose_workflow`."""

    def test_override(self) -> None:
        changes = Changes(string="Distribution: unstable\n")
        profile = {
            "debusine_workflows_by_distribution": {
                "unstable": "upload-to-unstable"
            },
            "debusine_workflow": "overridden",
        }

        self.assertEqual(choose_workflow(changes, profile), "overridden")

    def test_select_from_changes(self) -> None:
        changes = Changes(string="Distribution: unstable\n")
        profile = {
            "debusine_workflows_by_distribution": {
                "experimental": "upload-to-experimental",
                "unstable": "upload-to-unstable",
            }
        }

        self.assertEqual(
            choose_workflow(changes, profile), "upload-to-unstable"
        )

    def test_unknown(self) -> None:
        changes = Changes(string="Distribution: oldstable\n")
        profile = {
            "debusine_workflows_by_distribution": {
                "unstable": "upload-to-unstable"
            }
        }

        with self.assertRaisesRegex(
            NoWorkflowError,
            "No debusine workflow is configured for 'oldstable'",
        ):
            choose_workflow(changes, profile)

    def test_wildcard_fallback(self) -> None:
        changes = Changes(string="Distribution: trixie\n")
        profile = {
            "debusine_workflows_by_distribution": {
                "unstable": "upload-to-unstable",
                "*": "upload-to-{distribution}",
            }
        }

        self.assertEqual(choose_workflow(changes, profile), "upload-to-trixie")


class TestCheckWorkflow(TestCase):
    """Test the pre-upload hook to check that we have a workflow."""

    def test_success(self) -> None:
        changes = Changes(string="Distribution: unstable\n")
        profile = {
            "debusine_workflows_by_distribution": {
                "unstable": "upload-to-unstable"
            }
        }

        check_workflow(changes, profile, FakeInterface())

    def test_failure(self) -> None:
        changes = Changes(string="Distribution: unknown\n")
        profile = {
            "debusine_workflows_by_distribution": {
                "unstable": "upload-to-unstable"
            }
        }

        with self.assertRaisesRegex(
            NoWorkflowError,
            "No debusine workflow is configured for 'unknown'",
        ) as raised:
            check_workflow(changes, profile, FakeInterface())
        self.assertIsInstance(raised.exception, HookException)


class TestCreateWorkflow(TestCase):
    """Test the post-upload hook to create a workflow."""

    def test_success(self) -> None:
        changes = Changes(string="Distribution: unstable\n")
        profile = load_local_profile("debusine.debian.net")
        profile["debusine_workspace"] = "base"
        profile["debusine_artifact_ids"] = {"changes": 42}

        with (
            mock_client_config(
                "https://debusine.example.net/api", "some-token"
            ),
            mock.patch(
                "debusine.client.debusine.Debusine.workflow_create",
                return_value=create_work_request_response(id=123),
            ) as mock_workflow_create,
            self.assertLogsContains(
                "Created workflow: "
                "https://debusine.example.net/debian/base/work-request/123/",
                logger="dput",
                level=logging.INFO,
            ),
        ):
            create_workflow(changes, profile, FakeInterface())

        mock_workflow_create.assert_called_once_with(
            CreateWorkflowRequest(
                template_name="upload-to-unstable",
                workspace="base",
                task_data={"source_artifact": 42},
            )
        )

    def test_no_workflow(self) -> None:
        changes = Changes(string="Distribution: unknown\n")
        profile = load_local_profile("debusine.debian.net")
        # Disable fallback configured in this profile
        del profile["debusine_workflows_by_distribution"]["*"]

        with mock.patch(
            "debusine.client.debusine.Debusine.workflow_create"
        ) as mock_workflow_create:
            create_workflow(changes, profile, FakeInterface())

        mock_workflow_create.assert_not_called()

    def test_non_default_workflow_data(self) -> None:
        changes = Changes(string="Distribution: unstable\n")
        profile = load_local_profile("debusine.debian.net")
        profile["debusine_workspace"] = "base"
        # The value below replaces the @UPLOAD@ special value from
        # the above profile
        profile["debusine_artifact_ids"] = {"changes": 42}
        profile["debusine_workflow_data"]["enable_piuparts"] = False
        # We test the @DISTRIBUTION@ special value at the same time
        profile["debusine_workflow_data"]["codename"] = "@DISTRIBUTION@"

        with (
            mock_client_config(
                "https://debusine.example.net/api", "some-token"
            ),
            mock.patch(
                "debusine.client.debusine.Debusine.workflow_create",
                return_value=create_work_request_response(id=123),
            ) as mock_workflow_create,
            self.assertLogsContains(
                "Created workflow: "
                "https://debusine.example.net/debian/base/work-request/123/",
                logger="dput",
                level=logging.INFO,
            ),
        ):
            create_workflow(changes, profile, FakeInterface())

        mock_workflow_create.assert_called_once_with(
            CreateWorkflowRequest(
                template_name="upload-to-unstable",
                workspace="base",
                task_data={
                    "source_artifact": 42,
                    "enable_piuparts": False,
                    "codename": "unstable",
                },
            )
        )

    def test_workflow_create_error(self) -> None:
        """
        Exceptions from ``workflow_create`` are wrapped in a HookException.

        This is handled cleanly by the ``dput`` command line.
        """
        changes = Changes(string="Distribution: unstable\n")
        profile = load_local_profile("debusine.debian.net")
        profile["debusine_workspace"] = "base"
        profile["debusine_artifact_ids"] = {"changes": 42}
        error = DebusineError(
            title="Something went wrong",
            detail="Something went badly wrong",
        )

        with (
            mock_client_config(
                "https://debusine.example.net/api", "some-token"
            ),
            mock.patch(
                "debusine.client.debusine.Debusine.workflow_create",
                side_effect=error,
            ),
            self.assertRaises(HookException) as raised,
        ):
            create_workflow(changes, profile, FakeInterface())

        self.assertIsInstance(raised.exception, HookException)
        self.assertEqual(
            str(raised.exception),
            f"Failed to create workflow upload-to-unstable: {error}",
        )
        self.assertEqual(raised.exception.__cause__, error)



# Copyright 2020-2023 Enrico Zini <enrico@debian.org>
# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Views needed to interact with external authentication providers.

Login and logout hooks are implemented as mixing for the normal
django.contrib.auth.LoginView/LogoutView.
"""

import logging
from collections.abc import Sequence
from typing import Any

from django import http
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.core.handlers.exception import response_for_exception
from django.db import transaction
from django.http import HttpRequest, HttpResponse, HttpResponseBase
from django.shortcuts import redirect
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.generic import View
from prometheus_client import Counter

from debusine.db.models import Identity
from debusine.server.signon import providers
from debusine.server.signon.middleware import RequestSignonProtocol

log = logging.getLogger("debusine.server.signon")

signon_oidc_callbacks = Counter(
    "debusine_signon_oidc_callbacks",
    "OpenID Connect authentication callbacks",
    labelnames=("issuer", "activated"),
)
for provider in getattr(settings, "SIGNON_PROVIDERS", ()):  # pragma: no cover
    signon_oidc_callbacks.labels(issuer=provider.name, activated=True).reset()
    signon_oidc_callbacks.labels(issuer=provider.name, activated=False).reset()


class SignonLogoutMixin:
    """Mixin to logout external signon providers in a logout view."""

    @method_decorator(never_cache)
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        """Wrap the normal logout to also log out identifiers."""
        if signon := getattr(request, "signon", None):
            signon.logout_identities()
        assert isinstance(self, View)
        return super().dispatch(request, *args, **kwargs)


class BindIdentityView(View):
    """
    Bind an external identity to the current user.

    This will initiate an external authentication, setting things up so that on
    success the identity is bound to the current user
    """

    def get(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """Check, setup, and redirect to the external identity provider."""
        if not request.user.is_authenticated:
            raise PermissionDenied

        try:
            provider = providers.get(self.kwargs["name"])
        except ImproperlyConfigured:
            raise PermissionDenied
        bound_provider = provider.bind(request)
        url = bound_provider.get_authorization_url("bind")
        return redirect(url)


class OIDCAuthenticationCallbackView(View):
    """
    Handle a callback from an external OIDC authentication provider.

    If successful, this activates the identity related to the provider,
    creating it if missing
    """

    def _validate(
        self, request: http.HttpRequest
    ) -> tuple[dict[str, Any], Sequence[str]]:
        """
        Validate the information from the remote OIDC provider.

        :return: the claims dict and the options passed to
                 BoundProvider.get_authorization_url
        """
        name = self.kwargs["name"]
        try:
            provider = providers.get(name)
        except ImproperlyConfigured:
            raise http.Http404

        bound_provider = provider.bind(request)
        try:
            bound_provider.load_tokens()
        except providers.OIDCValidationError as e:
            log.info(  # noqa: G200
                "OIDC validation for provider %r failed: %s", name, str(e)
            )
            raise PermissionDenied

        return bound_provider.id_token_claims, bound_provider.options

    def get(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """
        Notify successful authentication from from the external OIDC server.

        This is called by the external OIDC server.

        Validate the server information, activate the relevant Identity and
        recompute authentication information with the new information.
        """
        name = self.kwargs["name"]
        try:
            claims, options = self._validate(request)
        except Exception:
            signon_oidc_callbacks.labels(issuer=name, activated=False).inc()
            raise

        try:
            identity = Identity.objects.get(issuer=name, subject=claims["sub"])
        except Identity.DoesNotExist:
            identity = Identity.objects.create(
                issuer=name,
                subject=claims["sub"],
            )

        # Remove the audience claim, which we don't need to store
        claims.pop("aud", None)

        identity.claims = claims
        identity.last_used = timezone.now()
        identity.save()

        with transaction.atomic():
            # Handle the exception ourselves, since we want to save the
            # identity's last_used state even if this fails.
            try:
                assert isinstance(request, RequestSignonProtocol)
                activated = request.signon.activate_identity(identity, *options)
            except Exception as exc:
                signon_oidc_callbacks.labels(issuer=name, activated=False).inc()
                return response_for_exception(request, exc)
            else:
                signon_oidc_callbacks.labels(
                    issuer=name, activated=activated
                ).inc()

        if not (next_url := request.session.pop("sso_callback_next_url", None)):
            next_url = getattr(
                settings, "SIGNON_DEFAULT_REDIRECT", "homepage:homepage"
            )

        return redirect(next_url)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""WorkflowTemplate task_data gets renamed, preparing for runtime_parameters."""

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0031_rename_more_arch_all_build_architecture_and_suite'),
    ]

    operations = [
        migrations.RenameField(
            model_name='workflowtemplate',
            old_name='task_data',
            new_name='static_parameters',
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, workflow management commands."""

import argparse
from typing import override

from debusine.client.commands.base import (
    RequiredInputDataCommand,
    WorkspaceCommand,
)
from debusine.client.commands.work_requests import WorkRequestModelCommand
from debusine.client.models import CreateWorkflowRequest


class Start(
    RequiredInputDataCommand,
    WorkspaceCommand,
    WorkRequestModelCommand,
    group="workflow",
):
    """Start a workflow."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "name",
            type=str,
            help=(
                "Name of the workflow template (or task, with --no-template)"
                " used to instantiate the workflow"
            ),
        )
        parser.add_argument(
            "--no-template",
            action="store_false",
            dest="use_template",
            default=True,
            help=(
                "Start a workflow directly without a template"
                " (only allowed for workspace owners)"
            ),
        )

    @override
    def run(self) -> None:
        if self.args.use_template:
            workflow = CreateWorkflowRequest(
                template_name=self.args.name,
                workspace=self.workspace,
                task_data=self.input_data,
            )
        else:
            workflow = CreateWorkflowRequest(
                task_name=self.args.name,
                workspace=self.workspace,
                task_data=self.input_data,
            )

        with self._api_call_or_fail():
            workflow_created = self.debusine.workflow_create(workflow)

        self.feedback("Workflow started:")
        self.show(workflow_created)


class LegacyStart(
    Start, name="create-workflow", deprecated="see `workflow start`"
):
    """Create a workflow from a template."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the service status view."""

from typing import Any, ClassVar, override

from django.test import override_settings
from django.urls import reverse

from debusine.db.models import Artifact, WorkRequest, Worker
from debusine.tasks.models import WorkerType
from debusine.test.django import TestCase


class ServiceStatusViewTests(TestCase):
    """Tests for ServiceStatusView."""

    source: ClassVar[Artifact]
    environment: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source = cls.playground.create_minimal_source_package_artifact()
        environment = cls.playground.create_debian_environment()
        assert environment.artifact is not None
        cls.environment = environment.artifact

    def create_worker_task(
        self,
        architecture: str,
        *,
        mark_pending: bool = False,
        mark_running: bool = False,
        worker: Worker | None = None,
    ) -> WorkRequest:
        """Create a worker task for the given architecture."""
        work_request = self.playground.create_sbuild_work_request(
            source=self.source,
            environment=self.environment,
            architecture=architecture,
        )
        self.playground.advance_work_request(
            work_request,
            mark_pending=mark_pending,
            mark_running=mark_running,
            worker=worker,
        )
        return work_request

    def assertServiceStatus(self, expected: dict[str, Any]) -> None:
        """Assert that the service-status API response matches expected."""
        response = self.client.get(reverse("api:service-status"))
        response_data = response.json()
        self.assertEqual(response_data, expected)

    def test_output_bland(self) -> None:
        """Test output when there's no data to report."""
        self.assertServiceStatus(
            {
                "external_workers_arch": {},
                "external_workers_native_arch": {},
                "tasks": {
                    "Internal": {"pending": 0, "running": 0},
                    "Server": {"pending": 0, "running": 0},
                    "Signing": {"pending": 0, "running": 0},
                    "Wait": {"pending": 0, "running": 0},
                    "Worker": {"pending": 0, "running": 0},
                    "Workflow": {"pending": 0, "running": 0},
                },
                "worker_tasks": {
                    "celery": {"pending": 0, "running": 0},
                    "external": {"pending": 0, "running": 0},
                    "signing": {"pending": 0, "running": 0},
                },
                "worker_tasks_arch": {},
                "workers": {
                    "celery": {
                        "busy": 0,
                        "connected": 0,
                        "idle": 0,
                        "registered": 0,
                    },
                    "external": {
                        "busy": 0,
                        "connected": 0,
                        "idle": 0,
                        "registered": 0,
                    },
                    "signing": {
                        "busy": 0,
                        "connected": 0,
                        "idle": 0,
                        "registered": 0,
                    },
                },
            }
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_output_example(self) -> None:
        """Test output when there's no data to report."""
        # 1 connected amd64 worker
        worker_amd64_1 = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64", "i386"],
            }
        )
        worker_amd64_1.mark_connected()

        # 2 arm64 workers, 1 connected
        self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arm64",
                "system:architectures": ["arm64"],
            }
        ).mark_connected()
        self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arm64",
                "system:architectures": ["arm64"],
            }
        )
        # 1 celery worker
        self.playground.create_worker(
            worker_type=WorkerType.CELERY,
        ).mark_connected()
        # 1 signing worker
        self.playground.create_worker(
            worker_type=WorkerType.SIGNING,
        ).mark_connected()

        self.create_worker_task("amd64", mark_pending=True)
        self.create_worker_task(
            "amd64", mark_running=True, worker=worker_amd64_1
        )
        self.create_worker_task("arm64", mark_pending=True)

        self.assertServiceStatus(
            {
                "external_workers_arch": {
                    "amd64": {
                        "busy": 1,
                        "connected": 1,
                        "idle": 0,
                        "registered": 1,
                    },
                    "arm64": {
                        "busy": 0,
                        "connected": 1,
                        "idle": 1,
                        "registered": 2,
                    },
                    "i386": {
                        "busy": 1,
                        "connected": 1,
                        "idle": 0,
                        "registered": 1,
                    },
                },
                "external_workers_native_arch": {
                    "amd64": {
                        "busy": 1,
                        "connected": 1,
                        "idle": 0,
                        "registered": 1,
                    },
                    "arm64": {
                        "busy": 0,
                        "connected": 1,
                        "idle": 1,
                        "registered": 2,
                    },
                },
                "tasks": {
                    "Internal": {"pending": 0, "running": 0},
                    "Server": {"pending": 0, "running": 0},
                    "Signing": {"pending": 0, "running": 0},
                    "Wait": {"pending": 0, "running": 0},
                    "Worker": {"pending": 2, "running": 1},
                    "Workflow": {"pending": 0, "running": 0},
                },
                "worker_tasks": {
                    "celery": {"pending": 0, "running": 0},
                    "external": {"pending": 2, "running": 1},
                    "signing": {"pending": 0, "running": 0},
                },
                "worker_tasks_arch": {
                    "amd64": {"pending": 1, "running": 1},
                    "arm64": {"pending": 1, "running": 0},
                    "i386": {"pending": 0, "running": 0},
                },
                "workers": {
                    "celery": {
                        "busy": 0,
                        "connected": 1,
                        "idle": 1,
                        "registered": 1,
                    },
                    "external": {
                        "busy": 1,
                        "connected": 2,
                        "idle": 1,
                        "registered": 3,
                    },
                    "signing": {
                        "busy": 0,
                        "connected": 1,
                        "idle": 1,
                        "registered": 1,
                    },
                },
            }
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test decorated ``migrate`` command."""

from unittest import mock

from django.core.management import CommandError
from django.db.utils import DatabaseError

from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


class MigrateCommandTests(TestCase):
    """Tests for the ``migrate`` command."""

    def test_operational_error_is_caught(self) -> None:
        """Test that OperationalError raised in execute is dealt with."""
        with (
            mock.patch(
                "django.core.management.BaseCommand.execute",
                side_effect=DatabaseError("Boom"),
            ),
            self.assertRaisesRegex(
                CommandError, r"^Database error: Boom$"
            ) as raised,
        ):
            call_command("migrate")

        self.assertEqual(raised.exception.returncode, 3)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Helpers for testing signing tasks."""

from pathlib import Path
from typing import Any, override

from django.utils import timezone

from debusine.artifacts import LocalArtifact
from debusine.client.models import (
    ArtifactResponse,
    FileResponse,
    FilesResponseType,
)
from debusine.signing.tasks import BaseSigningTask
from debusine.tasks.tests.helpers import ExternalTaskTestCase
from debusine.test.test_utils import create_artifact_response
from debusine.utils import calculate_hash


class SigningTaskTestCase[TaskClass: BaseSigningTask[Any, Any]](
    ExternalTaskTestCase[TaskClass]
):
    """Helper mixin for Task tests."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.example_url = "http://example.org/"

    def make_files_response(self, files: dict[str, Path]) -> FilesResponseType:
        """Make a fake files response."""
        return FilesResponseType(
            {
                name: FileResponse(
                    size=path.stat().st_size,
                    checksums={"sha256": calculate_hash(path, "sha256").hex()},
                    type="file",
                    url=self.example_url,
                )
                for name, path in files.items()
            }
        )

    def make_artifact_response(
        self, artifact: LocalArtifact[Any], artifact_id: int
    ) -> ArtifactResponse:
        """Make a fake artifact response."""
        return create_artifact_response(
            id=artifact_id,
            workspace="System",
            category=artifact.category,
            created_at=timezone.now(),
            data=artifact.data.model_dump(),
            download_tar_gz_url=self.example_url,
            files_to_upload=[],
            files=self.make_files_response(artifact.files),
        )
